
## Copyright (c) 2009-2010, Scott D. Peckham
## May, July, August 2009
## May 2010 (changes to unit_test() and read_cfg_file()

#-----------------------------------------------------------------------
#  NOTES:  This file defines an Energy-Balance ET component
#          and related functions.  It inherits from the ET
#          "base class" in "ET_base.py".
#-----------------------------------------------------------------------
#  NB!     Be sure to incorporate Bob Bolton's Nov 5, 2009
#          revisions ASAP!
#-----------------------------------------------------------------------
#
#  unit_test()
#
#  class ET_energy_balance
#
#      get_attribute()      # (10/26/11)
#      check_input_types()
#      update_ET_rate()

#  Functions:
#      Energy_Balance_ET_Rate

#-----------------------------------------------------------------------

from numpy import *
import numpy

import os

import cfg_files as cfg
import ET_base
import tf_utils

#-----------------------------------------------------------------------
def unit_test():

    ec = ET_component()
    ec.CCA   = False
    ec.DEBUG = False
    
    #-----------------------------------------
    # This function adjusts for the platform
    # and can be changed in "tf_utils.py".
    #-----------------------------------------
    cfg_directory   = tf_utils.TF_Test_Directory()
    os.chdir( cfg_directory )
    cfg_prefix      = tf_utils.TF_Test_Case_Prefix()
    ec.site_prefix  = 'Treynor'  #############  
  
    ec.run_model( cfg_directory=cfg_directory,
                  cfg_prefix=cfg_prefix )

##    ec.initialize( cfg_prefix=cfg_prefix, mode="driver" )
 
##    print 'nx =', ec.nx
##    print 'ny =', ec.ny
    
#   unit_test()
#-----------------------------------------------------------------------
class ET_component(ET_base.ET_component):

    #-------------------------------------------------------------------
    def get_attribute(self, att_name):

        map = {'comp_name':          'EvapEnergyBalance',
               'version':            '3.1',
               'model_name':         'Evaporation_Energy_Balance',
               'model_family':       'TopoFlow',
               'cfg_template_file':  'Evap_Energy_Balance.cfg.in',
               'cfg_extension':      '_evap_energy_balance.cfg',
               'cmt_var_prefix':     '/EvapEnergyBalance/Input/Var/',
               'gui_xml_file':       '/home/csdms/cca/topoflow/3.1/src/share/cmt/gui/Evap_Energy_Balance.xml',
               'dialog_title':       'Evaporation: Energy Balance Parameters',
               'time_step_type':     'fixed',
               'time_units':         'seconds',
               'mesh_type':          'none',
               'author_name':        'Scott Peckham'}

        try:
            return map[ att_name.lower() ]
        except:
            print '###################################################'
            print ' ERROR: Could not find attribute: ' + att_name
            print '###################################################'
            print ' '

    #   get_attribute() 
    #-------------------------------------------------------------------
    def check_input_types(self):

        #---------------------------------------------------
        # Note: h0_snow is used for law-of-wall roughness.
        #---------------------------------------------------

        #--------------------------------------------------------
        # As of 7/9/10, Qn_SW and Qn_LW are computed internally
        # from other vars, including slope and aspect grids.
        # So they'll always be grids and so will self.ET
        # unless PRECIP_ONLY = True.
        #--------------------------------------------------------
        # Need to get a boolean value, PRECIP_ONLY from the
        # "meteorology" component.  This doesn't work:
        # PRECIP_ONLY = self.get_port_data('PRECIP_ONLY', 'mp')
        # But the method just below does work.  Should we add
        # a new IRF port method called "get_boolean()" ??
        # Note that "self.mp" here is an embedded CCA port.
        #-------------------------------------------------------- 
        PRECIP_ONLY = self.mp.get_scalar_long('PRECIP_ONLY')
        Q_sum_IS_SCALAR = (PRECIP_ONLY == 1)
        if (self.DEBUG):
            print 'In ET component: PRECIP_ONLY =', PRECIP_ONLY
        ## Q_sum_IS_SCALAR = self.mp.is_scalar('Q_sum')
                         
        are_scalars = numpy.array([
                         self.is_scalar('T_soil_x'),
                         self.is_scalar('soil_x'),
                         self.is_scalar('K_soil'),
                         #----------------------------
                         self.sp.is_scalar('h_snow'),
                         #----------------------------
                         Q_sum_IS_SCALAR,
                         self.mp.is_scalar('Qe'),
                         self.mp.is_scalar('T_air') ])
        
        self.ALL_SCALARS = numpy.all(are_scalars)

        ## self.ALL_SCALARS = False
        
    #   check_input_types()
    #-------------------------------------------------------------------
    def update_ET_rate(self):

        #--------------------------------------------------------------
        # Notes: Qet   = energy used for ET of water from surface
        #        Qn_SW = net shortwave radiation flux (solar)
        #        Qn_LW = net longwave radiation flux (air, surface)
        #        Qh    = sensible heat flux from turbulent convection
        #                between snow surface and air
        #        Qc    = energy transferred from surface to subsurface

        #        All of the Q's have units of [W/m^2].

        #        T_air    = air temperature [deg_C]
        #        T_surf   = soil temp at the surface [deg_C]
        #        T_soil_x = soil temp at depth of x meters [deg_C]

        #        K_soil = thermal conductivity of soil [W/m/deg_C]
        #        K_soil = 0.45   ;[W/m/deg_C] (thawed soil; moisture
        #                     content near field capacity)
        #        K_soil = 1.0    ;[W/m/deg_C] (frozen soil)

        #        z0_air = roughness length scale [m]
        #        h_snow = snow depth [m]
        #--------------------------------------------------------------
        # NB!  h_snow is needed by the Bulk_Exchange_Coeff function
        #      to adjust reference height, z.
        #--------------------------------------------------------------
        Q_sum  = self.get_port_data('Q_sum',  self.mp)
        Qe     = self.get_port_data('Qe',     self.mp)
        T_surf = self.get_port_data('T_surf', self.mp)

        ################################################
        #  START USING "update_Qc" in "ET_base.py" ??
        ################################################
        
        #---------------------------------------------
        # Compute the conductive energy between the
        # surface and subsurface using Fourier's law
        #---------------------------------------------
        delta_T = (self.T_soil_x - T_surf)
        Qc      = self.K_soil * delta_T / self.soil_x
        
        #-------------------------------------------
        # Compute energy available for evaporation
        #-------------------------------------------
        # self.Qet = (Qn_SW + Qn_LW + Qh + Qc)
        Qet = (Q_sum - Qe)
        ################################################
        #  DO WE NEED TO SUBTRACT Qe HERE ??
        ################################################
        
        #------------------------------------------
        # Lf = latent heat of fusion [J/kg]
        # Lv = latent heat of vaporization [J/kg]
        # ET = (Qet / (rho_w * Lv))
        #------------------------------------------
        # rho_w = 1000d       ;[kg/m^3]
        # Lv    = -2500000d   ;[J/kg]
        # So (rho_w * Lv) = -2.5e+9  [J/m^3]
        #------------------------------------------
        # ET is a loss, but returned as positive.
        #------------------------------------------
        ET = (Qet / float64(2.5E+9))  # [m/s]
        self.ET = maximum(ET, float64(0))
    
    #   update_ET_rate()
    #-------------------------------------------------------------------

#-----------------------------------------------------------------------
#-----------------------------------------------------------------------
##def Energy_Balance_ET_Rate(K_soil, T_soil_x, soil_x, \
##                           Qn_SW, Qn_LW, T_air, T_surf, \
##                           uz, z, z0_air, rho_air, \
##                           Cp_air, h_snow):
##
##    #--------------------------------------------------------------
##    # Notes: Qet   = energy used for ET of water from surface
##    #        Qn_SW = net shortwave radiation flux (solar)
##    #        Qn_LW = net longwave radiation flux (air, surface)
##    #        Qh    = sensible heat flux from turbulent convection
##    #                between snow surface and air
##    #        Qc    = energy transferred from surface to subsurface
##
##    #        All of the Q's have units of [W/m^2].
##
##    #        T_air    = air temperature [deg_C]
##    #        T_surf   = soil temp at the surface [deg_C]
##    #        T_soil_x = soil temp at depth of x meters [deg_C]
##
##    #        K_soil = thermal conductivity of soil [W/m/deg_C]
##    #        K_soil = 0.45   ;[W/m/deg_C] (thawed soil; moisture
##    #                     content near field capacity)
##    #        K_soil = 1.0    ;[W/m/deg_C] (frozen soil)
##
##    #        z0_air = roughness length scale [m]
##    #        h_snow = snow depth [m]
##    #--------------------------------------------------------------
##    # NB!  h_snow is needed by the Bulk_Exchange_Coeff function
##    #      to adjust reference height, z.  It is not a pointer.
##    #--------------------------------------------------------------
##    
##    #--------------------------------
##    # Get bulk exchange coefficient
##    #--------------------------------
##    Dh = Bulk_Exchange_Coeff(uz, z, h_snow, z0_air, T_air, T_surf)
##    
##    #-----------------------------
##    # Compute sensible heat flux
##    #-----------------------------
##    #** T_surf = T0  ;?????????
##    Qh = Sensible_Heat_Flux(rho_air, Cp_air, Dh, T_air, T_surf)
##    #Formula:  Qh = (rho_air * Cp_air) * Dh * (T_air - T_surf)
##    
##    #---------------------------------------------
##    # Compute the conductive energy between the
##    # surface and subsurface using Fourier's law
##    #---------------------------------------------
##    Qc = K_soil * (T_soil_x - T_surf) / (soil_x)
##    
##    #--------------------------------
##    # Qn_SW and Qn_LW are pointers,
##    # others are local variables
##    #--------------------------------
##    Qnet = (Qn_SW + Qn_LW)
##    Qet  = (Qnet + Qh + Qc)
##    
##    #------------------------------------------
##    # Lf = latent heat of fusion [J/kg]
##    # Lv = latent heat of vaporization [J/kg]
##    # ET = (Qet / (rho_w * Lv))
##    #------------------------------------------
##    # rho_w = 1000d       ;[kg/m^3]
##    # Lv    = -2500000d   ;[J/kg]
##    # So (rho_w * Lv) = -2.5e+9  [J/m^3]
##    #-------------------------------------
##    ET = (Qet / float32(2.5E+9))  #[m/s]  (A loss, but returned as positive.)
##    
##    return maximum(ET, float64(0))
##    
###   Energy_Balance_ET_Rate
#-----------------------------------------------------------------------

