
## Copyright (c) 2009-2010, Scott D. Peckham
## January 2009  (converted from IDL)
## May, July, August 2009
## May 2010 (changes to unit_test() and read_cfg_file()

#################################################
#  NB!  See note at "THIS MAY BE COSTLY"
#################################################

#-----------------------------------------------------------------------
#  NOTES:  This file defines a Priestley-Taylor ET component
#          and related functions.  It inherits from the ET
#          "base class" in "ET_base.py".
#-----------------------------------------------------------------------
#
#  unit_test()
#
#  class ET_component
#
#      get_attribute()         # (10/26/11)
#      check_input_types()
#      update_ET_rate()
#
#  Functions:
#      Priestley_Taylor_ET_Rate
#
#-----------------------------------------------------------------------

from numpy import *
import numpy

import os

import cfg_files as cfg
import ET_base
import tf_utils

#-----------------------------------------------------------------------
def unit_test():

    ec = ET_component()
    ec.CCA   = False
    ec.DEBUG = False
    
    #-----------------------------------------
    # This function adjusts for the platform
    # and can be changed in "tf_utils.py".
    #-----------------------------------------
    cfg_directory   = tf_utils.TF_Test_Directory()
    os.chdir( cfg_directory )
    cfg_prefix      = tf_utils.TF_Test_Case_Prefix()
    ec.site_prefix  = 'Treynor'  #############  
  
    ec.run_model( cfg_directory=cfg_directory,
                  cfg_prefix=cfg_prefix )

##    ec.initialize( cfg_prefix=cfg_prefix, mode="driver" )
 
##    print 'nx =', ec.nx
##    print 'ny =', ec.ny
    
#   unit_test()
#-----------------------------------------------------------------------
class ET_component(ET_base.ET_component):

    #-------------------------------------------------------------------
    def get_attribute(self, att_name):

        map = {'comp_name':          'EvapPriestleyTaylor',
               'version':            '3.1',
               'model_name':         'Evaporation_Priestley_Taylor',
               'model_family':       'TopoFlow',
               'cfg_template_file':  'Evap_Priestley_Taylor.cfg.in',
               'cfg_extension':      '_evap_priestley_taylor.cfg',
               'cmt_var_prefix':     '/EvapPriestleyTaylor/Input/Var/',
               'gui_xml_file':       '/home/csdms/cca/topoflow/3.1/src/share/cmt/gui/Evap_Priestley_Taylor.xml',
               'dialog_title':       'Evaporation: Priestley-Taylor Parameters',
               'time_step_type':     'fixed',
               'time_units':         'seconds',
               'mesh_type':          'none',
               'author_name':        'Scott Peckham'}

        try:
            return map[ att_name.lower() ]
        except:
            print '###################################################'
            print ' ERROR: Could not find attribute: ' + att_name
            print '###################################################'
            print ' '

    #   get_attribute() 
    #-------------------------------------------------------------------
    def check_input_types(self):

        #--------------------------------------------------------
        # As of 7/9/10, Qn_SW and Qn_LW are computed internally
        # from other vars, including slope and aspect grids.
        # So they'll always be grids and so will self.ET
        # unless PRECIP_ONLY = True.
        #--------------------------------------------------------
        # Need to get a boolean value, PRECIP_ONLY from the
        # "meteorology" component.  This doesn't work:
        # PRECIP_ONLY = self.get_port_data('PRECIP_ONLY', 'mp')
        # But the method just below does work.  Should we add
        # a new IRF port method called "get_boolean()" ??
        # Note that "self.mp" here is an embedded CCA port.
        #--------------------------------------------------------        
        PRECIP_ONLY = self.mp.get_scalar_long('PRECIP_ONLY')
        Qn_SW_IS_SCALAR = (PRECIP_ONLY == 1)
        Qn_LW_IS_SCALAR = (PRECIP_ONLY == 1)
        if (self.DEBUG):
            print 'In ET component: PRECIP_ONLY =', PRECIP_ONLY
        ## Qn_SW_IS_SCALAR = self.mp.is_scalar('Qn_SW'),
        ## Qn_LW_IS_SCALAR = self.mp.is_scalar('Qn_LW'),
        
        are_scalars = numpy.array([
                         self.is_scalar('T_soil_x'),
                         self.is_scalar('soil_x'),
                         self.is_scalar('K_soil'),
                         self.is_scalar('alpha'),
                         #-------------------------------
                         Qn_SW_IS_SCALAR,
                         Qn_LW_IS_SCALAR,
                         self.mp.is_scalar('T_air'),
                         self.mp.is_scalar('T_surf'),
                         #-------------------------------
                         self.cp.is_scalar('d'),
                         #-------------------------------
                         self.gp.is_scalar('h_table') ])

        self.ALL_SCALARS = numpy.all(are_scalars)

        ## self.ALL_SCALARS = False
        
    #   check_input_types()
    #-------------------------------------------------------------------
    def update_ET_rate(self):

        #--------------------------------------------------------------
        # Notes: Qet   = energy used for ET of water from surface
        #        Qn_SW = net shortwave radiation flux (solar)
        #        Qn_LW = net longwave radiation flux (air, surface)
        #        Qh    = sensible heat flux from turbulent convection
        #                between snow surface and air
        #        Qc    = energy transferred from surface to subsurface

        #        All of the Q's have units of [W/m^2].

        #        T_air    = air temperature [deg_C]
        #        T_surf   = soil temp at the surface [deg_C]
        #        T_soil_x = soil temp at depth of x meters [deg_C]

        #        Ks   = thermal conductivity of soil [W/m/deg_C]
        #        Ks = 0.45   ;[W/m/deg_C] (thawed soil; moisture content
        #                     near field capacity)
        #        Ks = 1.0    ;[W/m/deg_C] (frozen soil)

        #        alpha = evaporation parameter
        #        alpha = 0.95   ;(average found by Rouse)
        #        alpha = 1.26   ;(Jackson et al. (1996), at saturation)

        #        Modification of alpha:  alpha = (a1 * R) + a2
        #        R  = 1.0d   ;(equals 1 for saturation; R in [0,1])
        #        a1 = 1.0d   ;(accounts for moisture content of soil)
        #        a2 = 0.2d   ;(accounts for vegetation effect)
        #--------------------------------------------------------------
        Qn_SW  = self.get_port_data('Qn_SW',  self.mp)
        Qn_LW  = self.get_port_data('Qn_LW',  self.mp)
        T_air  = self.get_port_data('T_air',  self.mp)
        T_surf = self.get_port_data('T_surf', self.mp)
        Q_net  = Qn_SW + Qn_LW
        
        #---------------------------------------------
        # Compute the conductive energy between the
        # surface and subsurface using Fourier's law
        #---------------------------------------------
        # soil_x is converted from [cm] to [m] when
        # it is read from the GUI and then stored
        #---------------------------------------------
        # In Qet formula, the constant 0.011 has
        # units of 1/[deg_C] to cancel T_air units.
        #---------------------------------------------  
        Qc  = self.K_soil * (self.T_soil_x - T_surf) / self.soil_x
        Qet = self.alpha * (float64(0.406) + (float64(0.011) * T_air)) * (Q_net - Qc)
        
        #-----------------------------------
        # Convert ET energy to a loss rate
        #------------------------------------------
        # Lf = latent heat of fusion [J/kg]
        # Lv = latent heat of vaporization [J/kg]
        # ET = (Qet / (rho_w * Lv))
        #------------------------------------------
        # rho_w = 1000d       ;[kg/m^3]
        # Lv    = -2500000d   ;[J/kg]
        # So (rho_w * Lv) = -2.5e+9  [J/m^3]
        #-------------------------------------
        ET = (Qet / float64(2.5E+9))  #[m/s]  (A loss, but returned as positive.)
        
        self.ET = numpy.maximum(ET, float64(0))

        ##########################################
        #  THIS MAY BE COSTLY.  BETTER WAY OR
        #  ALLOW ET TO BE A SCALAR ??
        ##########################################
        if (size(self.ET) == 1):
            self.ET += zeros((self.ny, self.nx), dtype='Float64')
            
    #   update_ET_rate()        
    #-------------------------------------------------------------------

#-----------------------------------------------------------------------
#-----------------------------------------------------------------------   
def Priestley_Taylor_ET_Rate(alpha, Ks, T_soil_x, soil_x, \
                             Qn_SW, Qn_LW, T_air, T_surf):

    #--------------------------------------------------------------
    # Notes: Qet   = energy used for ET of water from surface
    #        Qn_SW = net shortwave radiation flux (solar)
    #        Qn_LW = net longwave radiation flux (air, surface)
    #        Qh    = sensible heat flux from turbulent convection
    #                between snow surface and air
    #        Qc    = energy transferred from surface to subsurface

    #        All of the Q's have units of [W/m^2].

    #        T_air    = air temperature [deg_C]
    #        T_surf   = soil temp at the surface [deg_C]
    #        T_soil_x = soil temp at depth of x meters [deg_C]

    #        Ks   = thermal conductivity of soil [W/m/deg_C]
    #        Ks = 0.45   ;[W/m/deg_C] (thawed soil; moisture content
    #                     near field capacity)
    #        Ks = 1.0    ;[W/m/deg_C] (frozen soil)

    #        alpha = evaporation parameter
    #        alpha = 0.95   ;(average found by Rouse)
    #        alpha = 1.26   ;(Jackson et al. (1996), at saturation)

    #        Modification of alpha:  alpha = (a1 * R) + a2
    #        R  = 1.0d   ;(equals 1 for saturation; R in [0,1])
    #        a1 = 1.0d   ;(accounts for moisture content of soil)
    #        a2 = 0.2d   ;(accounts for vegetation effect)
    #--------------------------------------------------------------
    
    #---------------------------------------------
    # Compute the conductive energy between the
    # surface and subsurface using Fourier's law
    #---------------------------------------------
    # soil_x is converted from [cm] to [m] when
    # it is read from the GUI and then stored
    #---------------------------------------------
    # In Qet formula, the constant 0.011 has
    # units of 1/[deg_C] to cancel T_air units.
    #---------------------------------------------  
    Qc   = Ks * (T_soil_x - T_surf) / (soil_x)
    Qnet = Qn_SW + Qn_LW
    Qet  = alpha * (float64(0.406) + (float32(0.011) * T_air)) * (Qnet - Qc)
    
    #-----------------------------------
    # Convert ET energy to a loss rate
    #------------------------------------------
    # Lf = latent heat of fusion [J/kg]
    # Lv = latent heat of vaporization [J/kg]
    # ET = (Qet / (rho_w * Lv))
    #------------------------------------------
    # rho_w = 1000d       ;[kg/m^3]
    # Lv    = -2500000d   ;[J/kg]
    # So (rho_w * Lv) = -2.5e+9  [J/m^3]
    #-------------------------------------
    ET = (Qet / float32(2.5E+9))  #[m/s]  (A loss, but returned as positive.)
    
    return maximum(ET, float64(0))

#   Priestley_Taylor_ET_Rate
#-----------------------------------------------------------------------

