
## Copyright (c) 2009-2010, Scott D. Peckham
## January 2009  (converted from IDL)
## May, August 2009
## May 2010 (changes to initialize(), read_cfg_file(), etc.)

#-----------------------------------------------------------------------
#  NOTES:  This file defines a Smith-Parlange (3-parameter)
#          infiltration component and related functions.
#          It inherits from the infiltration "base class" in
#          "infil_base.py".
#-----------------------------------------------------------------------
#
#  unit_test()
#
#  class infil_richards_1D
#
#      get_attribute()              # (10/26/11)
#      initialize_layer_vars()
#      -------------------------
#      set_computed_input_vars()
#      check_input_types()
#      update_infil_rate()
#      -------------------------
#      open_input_files()
#      read_input_files()
#      close_input_files()
#      ----------------------
#      update_outfile_names()

#  Functions:
#      Smith_Parlange_Infil_Rate_v1
#      Smith_Parlange_Infil_Rate_1D
#      Smith_Parlange_Infil_Rate_3D
#      Smith_Parlange_Infil_Rate_v2  (uses previous 2)

#-----------------------------------------------------------------------

from numpy import *
import numpy

import os

import infil_base
import model_input
import tf_utils

#-----------------------------------------------------------------------
def unit_test():

    ic = infil_component()
    ic.CCA   = False
    ic.DEBUG = True

    #-----------------------------------------
    # This function adjusts for the platform
    # and can be changed in "tf_utils.py".
    #-----------------------------------------
    cfg_directory  = tf_utils.TF_Test_Directory()
    cfg_prefix     = tf_utils.TF_Test_Case_Prefix()
    ic.site_prefix = 'Treynor'  #############
    
    #-------------------------------
    # Initialize and 1 update call
    #-------------------------------
##    print 'STATUS =', ic.get_status()
##    ic.initialize( mode="driver" )
##    print 'STATUS =', ic.get_status()
##    time_sec = float64(0)
##    ic.update(time_sec)
##    print 'STATUS =', ic.get_status()

    #--------------------------------
    # Run model in stand-alone mode
    #--------------------------------
    ic.run_model( cfg_directory=cfg_directory,
                  cfg_prefix=cfg_prefix )

#   unit_test()
#-----------------------------------------------------------------------
class infil_component(infil_base.infil_component):

    #-------------------------------------------------------------------
    def get_attribute(self, att_name):

        map = {'comp_name':          'InfilSmithParlange',
               'version':            '3.1',
               'model_name':         'Infiltration_Smith_Parlange',
               'model_family':       'TopoFlow',
               'cfg_template_file':  'Infil_Smith_Parlange.cfg.in',
               'cfg_extension':      '_infil_smith_parlange.cfg',
               'cmt_var_prefix':     '/InfilSmithParlange/Input/Var/',
               'gui_xml_file':       '/home/csdms/cca/topoflow/3.1/src/share/cmt/gui/Infil_Smith_Parlange.xml',
               'dialog_title':       'Infiltration: Smith-Parlange Parameters',
               'time_step_type':     'fixed',
               'time_units':         'seconds',
               'mesh_type':          'none',
               'author_name':        'Scott Peckham'}

        try:
            return map[ att_name.lower() ]
        except:
            print '###################################################'
            print ' ERROR: Could not find attribute: ' + att_name
            print '###################################################'
            print ' '

    #   get_attribute() 
    #-------------------------------------------------------------------
    def initialize_layer_vars(self):

        #-------------------------------------------------------
        # Notes: We need to call initialize_layer_vars()
        #        before initialize_config_vars(), which may
        #        call read_cfg_file().  However, this means
        #        we haven't read "n_layers" yet, so just
        #        hardwire it here for now. (5/11/10)
        #-------------------------------------------------------
        n_layers = 1
        # n_layers = self.n_layers
        
        #-------------------------------------------------
        # Get arrays to store soil params for each layer
        #-------------------------------------------------
        self.soil_type = numpy.zeros([n_layers], dtype='|S100')
##        self.dz_val    = numpy.zeros([n_layers], dtype='Float64')    #### + dz3
##        self.nz_val    = numpy.zeros([n_layers], dtype='Int16')      #### + nz3
        #----------------------------------------------------------
        self.Ks_type  = numpy.zeros(n_layers, dtype='|S100')
        self.Ki_type  = numpy.zeros(n_layers, dtype='|S100')
        self.qs_type  = numpy.zeros(n_layers, dtype='|S100')
        self.qi_type  = numpy.zeros(n_layers, dtype='|S100')
        self.G_type   = numpy.zeros(n_layers, dtype='|S100')
        self.gam_type = numpy.zeros(n_layers, dtype='|S100')
        #--------------------------------------------------------        
        self.Ks_file  = numpy.zeros(n_layers, dtype='|S100')
        self.Ki_file  = numpy.zeros(n_layers, dtype='|S100')
        self.qs_file  = numpy.zeros(n_layers, dtype='|S100')
        self.qi_file  = numpy.zeros(n_layers, dtype='|S100')
        self.G_file   = numpy.zeros(n_layers, dtype='|S100')
        self.gam_file = numpy.zeros(n_layers, dtype='|S100')
        #---------------------------------------------------------
        # Note: self.Ks is a Python list.  Initially, each entry
        # is a numpy scalar (type 'numpy.float64').  However, we
        # can later change any list entry to a scalar or grid
        # (type 'numpy.ndarray'), according to its "Ks_type".
        #---------------------------------------------------------
        self.Ks  = list(numpy.zeros(n_layers, dtype='Float64'))
        self.Ki  = list(numpy.zeros(n_layers, dtype='Float64'))
        self.qs  = list(numpy.zeros(n_layers, dtype='Float64'))
        self.qi  = list(numpy.zeros(n_layers, dtype='Float64'))
        self.G   = list(numpy.zeros(n_layers, dtype='Float64'))
        self.gam = list(numpy.zeros(n_layers, dtype='Float64'))

    #   initialize_layer_vars()
    #-------------------------------------------------------------------
    def set_computed_input_vars(self):
         
        #--------------------------------------------------------
        # Define these here, so all components can use the same
        # output file functions, like "open_output_files()".
        #--------------------------------------------------------
        self.RICHARDS       = False
        self.SAVE_Q0_GRIDS  = False
        self.SAVE_ZW_GRIDS  = False
        self.SAVE_Q0_PIXELS = False
        self.SAVE_ZW_PIXELS = False
        
        #---------------------------------------------------------
        # Make sure that all "save_dts" are larger or equal to
        # the specified process dt.  There is no point in saving
        # results more often than they change.
        # Issue a message to this effect if any are smaller ??
        #---------------------------------------------------------
        self.save_grid_dt    = maximum(self.save_grid_dt,    self.dt)
        self.save_pixels_dt  = maximum(self.save_pixels_dt,  self.dt)

    #   set_computed_input_vars()   
    #-------------------------------------------------------------------
    def check_input_types(self):

        #------------------------------------------------------
        # Notes: Usually this will be overridden by a given
        #        method of computing ET.  But this one should
        #        work for Green-Ampt and Smith-Parlange.
        #------------------------------------------------------
        are_scalars = numpy.array([
                         ## self.pp.is_scalar('P'),
                         self.mp.is_scalar('P'),
                         self.sp.is_scalar('SM'),
                         self.gp.is_scalar('h_table'),
                         #------------------------------
                         self.is_scalar('Ks'),
                         self.is_scalar('Ki'),
                         self.is_scalar('qs'),
                         self.is_scalar('qi'),
                         self.is_scalar('G'),
                         self.is_scalar('gam')  ])

        self.ALL_SCALARS = numpy.all(are_scalars)
        
    #   check_input_types()
    #-------------------------------------------------------------------
    def update_infil_rate(self):

        #-----------------------------------------------------
        # This function is not totally correct but is stable
        # and gives results similar to the correct method.
        #-----------------------------------------------------
        Smith_Parlange_Infil_Rate_v1(self)

        #-----------------------------------------------------
        # Numerically more correct method (??) but unstable
        #-----------------------------------------------------
        ## r = self.P_total
        ## r_last = ???
        ## n = self.time_index
        ## Smith_Parlange_Infil_Rate_v2(self, r, r_last, n)
       
    #   update_infil_rate()
    #-------------------------------------------------------------------  
    def open_input_files(self):

        #-------------------------------------------------------
        # This method works for Green-Ampt and Smith-Parlange
        # but must be overridden for Richards 1D.
        #-------------------------------------------------------
        # NB! Green-Ampt and Smith-Parlange currently only
        #     support ONE layer (n_layers == 1).
        #-------------------------------------------------------
        self.Ks_unit  = []  # (empty lists to hold file objects)
        self.Ki_unit  = []
        self.qs_unit  = []
        self.qi_unit  = []
        self.G_unit   = []
        self.gam_unit = []

        for k in xrange(self.n_layers):
            self.Ks_file[k] = self.in_directory + self.Ks_file[k]
            self.Ki_file[k] = self.in_directory + self.Ki_file[k]
            self.qs_file[k] = self.in_directory + self.qs_file[k]
            self.qi_file[k] = self.in_directory + self.qi_file[k]
            self.G_file[k]  = self.in_directory + self.G_file[k]
            self.gam_file[k] = self.in_directory + self.gam_file[k]

            self.Ks_unit.append(  model_input.open_file(self.Ks_type[k],  self.Ks_file[k]) )
            self.Ki_unit.append(  model_input.open_file(self.Ki_type[k],  self.Ki_file[k]) )
            self.qs_unit.append(  model_input.open_file(self.qs_type[k],  self.qs_file[k]) )
            self.qi_unit.append(  model_input.open_file(self.qi_type[k],  self.qi_file[k]) )
            self.G_unit.append(   model_input.open_file(self.G_type[k],   self.G_file[k])  )
            self.gam_unit.append( model_input.open_file(self.gam_type[k], self.gam_file[k]) )

    #   open_input_files()
    #-------------------------------------------------------------------  
    def read_input_files(self):

        rti = self.rti

        #-------------------------------------------------------
        # All grids are assumed to have data type of Float32.
        #-------------------------------------------------------
        # This method works for Green-Ampt and Smith-Parlange
        # but must be overridden for Richards 1D.
        #-------------------------------------------------------
        # NB! Green-Ampt and Smith-Parlange currently only
        #     support ONE layer (n_layers == 1).
        #------------------------------------------------------- 
        for k in xrange(self.n_layers):
            Ks = model_input.read_next(self.Ks_unit[k], self.Ks_type[k], rti)
            if (Ks != None): self.Ks[k] = Ks

            Ki = model_input.read_next(self.Ki_unit[k], self.Ki_type[k], rti)
            if (Ki != None): self.Ki[k] = Ki

            qs = model_input.read_next(self.qs_unit[k], self.qs_type[k], rti)
            if (qs != None): self.qs[k] = qs

            qi = model_input.read_next(self.qi_unit[k], self.qi_type[k], rti)
            if (qi != None): self.qi[k] = qi
            
            G  = model_input.read_next(self.G_unit[k], self.G_type[k], rti)
            if (G != None): self.G[k] = G

            gam = model_input.read_next(self.gam_unit[k], self.gam_type[k], rti)
            if (gam != None): self.gam[k] = gam
          
    #   read_input_files()       
    #-------------------------------------------------------------------  
    def close_input_files(self):

        #-------------------------------------------------------
        # This method works for Green-Ampt and Smith-Parlange
        # but must be overridden for Richards 1D.
        #-------------------------------------------------------
        # NB! Green-Ampt and Smith-Parlange currently only
        #     support ONE layer (n_layers == 1).
        #-------------------------------------------------------        
        for k in xrange(self.n_layers):
            if (self.Ks_type[k]  != 'Scalar'): self.Ks_unit[k].close()        
            if (self.Ki_type[k]  != 'Scalar'): self.Ki_unit[k].close()
            if (self.qs_type[k]  != 'Scalar'): self.qs_unit[k].close()
            if (self.qi_type[k]  != 'Scalar'): self.qi_unit[k].close()
            if (self.G_type[k]   != 'Scalar'): self.G_unit[k].close()
            if (self.gam_type[k] != 'Scalar'): self.gam_unit[k].close()
            #------------------------------------------------------------
##            if (self.Ks_file[k]  != ''): self.Ks_unit[k].close()        
##            if (self.Ki_file[k]  != ''): self.Ki_unit[k].close()
##            if (self.qs_file[k]  != ''): self.qs_unit[k].close()
##            if (self.qi_file[k]  != ''): self.qi_unit[k].close()
##            if (self.G_file[k]   != ''): self.G_unit[k].close()
##            if (self.gam_file[k] != ''): self.gam_unit[k].close()
          
    #   close_input_files()
    #-------------------------------------------------------------------  
    def update_outfile_names(self):

        #-------------------------------------------------
        # Notes:  Append out_directory to outfile names.
        #-------------------------------------------------
        self.v0_gs_file = (self.out_directory + self.v0_gs_file)
        self.I_gs_file  = (self.out_directory + self.I_gs_file)
        #---------------------------------------------------------
        self.v0_ts_file = (self.out_directory + self.v0_ts_file)
        self.I_ts_file  = (self.out_directory + self.I_ts_file)

##        self.v0_gs_file = (self.case_prefix + '_2D-v0.rts')
##        self.I_gs_file  = (self.case_prefix + '_2D-I.rts')
##        self.q0_gs_file = (self.case_prefix + '_2D-q0.rts')
##        self.Zw_gs_file = (self.case_prefix + '_2D-Zw.rts')
##        #---------------------------------------------------------
##        self.v0_ts_file = (self.case_prefix + '_0D-v0.txt')
##        self.I_ts_file  = (self.case_prefix + '_0D-I.txt')
##        self.q0_ts_file = (self.case_prefix + '_0D-q0.txt')
##        self.Zw_ts_file = (self.case_prefix + '_0D-Zw.txt')

    #   update_outfile_names()   
    #-------------------------------------------------------------------  

#-----------------------------------------------------------------------
def Smith_Parlange_Infil_Rate_v1(self):

    #------------------------------------------------------------
    #Notes:  IN  = infiltration rate [m/s]
    #        Ks  = saturated hydraulic conductivity [m/s]
    #        Ki  = initial hydraulic conductivity [m/s]
    #        qs  = soil moisture content [dimless]
    #        qi  = soil initial moisture content [dimless]
    #         G  = capillary length scale [m]
    #         I  = cum. infiltration depth (since reset) [m]
    #         P  = precipitation rate [m/s]

    #        SM  = snowmelt rate [m/s]
    #         r  = (P + SM)  [m/s]

    #        Note that the infiltration rate has a max possible
    #        value of (P + SM) and asymptotes to Ks as the
    #        total infiltrated depth increases.  Need to reset
    #        this depth between "events", but haven't decided
    #        how to do this yet.

    #        Total infiltrated depth is incremented in the
    #        calling function, called Infiltration.
    #------------------------------------------------------------
    Ks  = self.Ks[0]   # (synonyms)
    Ki  = self.Ki[0]
    qs  = self.qs[0]
    qi  = self.qi[0]
    G   = self.G[0]
    gam = self.gam[0]
    
    dK = (Ks - Ki)
    dq = (qs - qi)
    t2 = exp((gam * self.I) / (G * dq)) - float64(1)
    IN = (gam * dK / t2) + Ks
    
    #---------------------------------------------------
    # Initially, IN = P_total, and all of the incoming
    # water infiltrates.  IN cannot exceed P_total.
    # Ponding time, Tp, is time until (IN < P_total).
    #---------------------------------------------------
    self.IN = minimum(IN, self.P_total)
    
    #-----------------------------------
    #Is P_total less than Ks anywhere ?
    #If so, set IN = P_total there.
    #-----------------------------------
    ## self.check_low_rainrate()  # (now done in infil_base.py)
    
#   Smith_Parlange_Infil_Rate_v1
#-----------------------------------------------------------------------
def Smith_Parlange_Infil_Rate_1D(self, r, r_last, n):

    #------------------------------------------------------------------
    #Notes:  This version was written on 5/8-11/06 using the method
    #        for dealing with variable rainfall that is described
    #        on p. 114-116 in Smith, R.E. (2002) Infiltration Theory
    #        for Hydrologic Applications, Water Resources Monograph
    #        15, AGU.  The method used previously gives very similar
    #        results, at least for sufficiently small time steps, but
    #        is technically incorrect.  Due to its simplicity, it may
    #        still be useful as a faster approximate method if we can
    #        estimate error for various parameter settings.

    #        The main difference with this version is that ponding
    #        times may occur between timesteps and two different
    #        cases are considered - ponding in the middle of a time
    #        step and ponding during a rise in rainrate at the very
    #        end of a timestep (MID_PONDING vs. END_PONDING). Also
    #        ponding time was not specifically computed before.

    #        In order to test the second case (END_PONDING) uniform

    #        precip was used with rates of 5e-5 m/s and 1e-4 m/s with
    #        durations of 8 minutes and 5 minutes.  The soil type of
    #        "silty clay" was selected to set infiltration parameters,
    #        the infiltration rate was set to 0.1 minutes (6 seconds,
    #        same as default for channel process).  The model was then
    #        run for 200 time steps and the infiltration rate was
    #        plotted (TEST_0D-v0.txt).  These settings triggered an
    #        END_PONDING event and results appeared to be correct.
    #        However, with these settings a virtually identical result
    #        was obtained by the much simpler incorrect method used by
    #        Green_AmptINfil_Rate_v1.  However, when infiltration
    #        time step was increased to 0.3 minutes (18 seconds),
    #        there was a fairly significant difference and had to
    #        change initial f from (r/2) to (0.4 * r) to avoid a jump
    #        in the infiltration curve.
    #------------------------------------------------------------------
    Ks  = self.Ks[0]   # (synonyms)
    Ki  = self.Ki[0]
    qs  = self.qs[0]
    qi  = self.qi[0]
    G   = self.G[0]
    gam = self.gam[0]
    
    #-----------------------------
    # If r <= Ks, ponding cannot
    # occur, so return f = r
    #-----------------------------
    if (r <= Ks):    
        return r
    
    #----------------------------------
    #Time at start and end of timestep
    #----------------------------------
    t_start = n * self.dt
    t_end = (n + 1) * self.dt
    
    #-----------------------
    #Other common variables
    #-----------------------
    dq  = (qs - qi)
    dK  = (Ks - Ki)
    fac = (G * dq / gam)
    
    PONDED = (self.tp != -float64(1))
    R_GT_KS = (r > Ks)
    if (not(PONDED) and R_GT_KS):    
        #--------------------------------------------
        #Compute test values of Ip  (Smith eqn 5.47)
        #--------------------------------------------
        #If (r_last lt Ks) then we'll have Ip1 < 0,
        #so ponding at end of interval can't occur.
        #But it seems it should be able to if r is
        #big enough or time step is long enough.
        #So set Ip1 large in this case ??
        #--------------------------------------------
        #*** if (r_last le *self.Ks) AND (n eq 1) then begin
        if (r_last <= Ks):    
            Ip1 = float64(999999)
        else:    
            
            Ip1 = fac * log(float64(1) + (gam * dK / (r_last - Ks)))
        Ip2 = fac * log(float64(1) + (gam * dK / (r - Ks)))
        dIp = (Ip2 - self.I)
        
        #------------------------------------------
        #Does ponding occur within this timestep ?
        #------------------------------------------
        #   Ia(n) = sum_{k=0}^n (r_k * dt)
        #   Ia(n-1) < Ip(r_n) < Ia(n)
        #   0       < [Ip(r_n) - Ia(n-1)] < r_n*dt
        #   I = Ia(n-1), not updated yet by caller
        #------------------------------------------
        MID_PONDING = logical_and((dIp > 0), (dIp < (r * self.dt)))
        if (MID_PONDING):    
            print 'MID_PONDING occurred.'
            #------------------------
            #For this case: Ip = Ip2
            #------------------------
            self.tp = t_start + (dIp / r)    #(r gt Ks, above)
            self.fp = r
        
        #----------------------------------------
        #Does ponding occur at end of timestep ?
        #----------------------------------------
        #   Ip(r_n) < Ia(n-1) < Ip(r_{n-1})
        #   [Ip(r_n) - Ia(n-1)] < 0  AND
        #    Ia(n-1) < Ip(r_{n-1})
        #----------------------------------------
        END_PONDING = logical_and(logical_and((dIp < 0), (self.I < Ip1)), logical_not(MID_PONDING))
        if (END_PONDING):    
            print 'END_PONDING occurred.'
            #--------------------------
            #For this case: Ip = *self.I
            #--------------------------
            self.tp = t_end
            self.fp = ((gam * dK) / (exp(self.I / fac) - float64(1))) + Ks
        
        #------------------------
        #For debugging & testing
        #------------------------
        if (logical_or(MID_PONDING, END_PONDING)):    
            TF_Print('tp = ' + TF_String(self.tp))
            TF_Print(' ')
        
    
    #--------------------------------------------------
    #If ponding still hasn't occurred, all "rain" will
    #infiltrate; infil rate = r.  RETURN to caller and
    #increment the cumulative depth, self.I by (r * dt)
    #--------------------------------------------------
    PONDED = (self.tp != -float64(1))
    if logical_not(PONDED):    
        return r
    
    #---------------------------------------------------
    #If ponding from the start or in previous 2 tests
    #then forge ahead and compute infiltration rate
    #via Newton-Raphson iteration using fp and tp.
    #---------------------------------------------------
    #First, set initial guess for iteration, f.
    #*self.fp was either set above or in a previous call.
    #---------------------------------------------------
    #   f = r / 2d    ;(converged usually,
    #                   but jumped to f < 0 sometimes)
    #   f = r         ;(converges, but jumps to f < 0)
    #   f  = 0.9 * r  ;(didn't converge)
    #   f = r / 100d  ;(didn't converge)
    #   f = 2d * r    ;(didn't converge)
    #   f = 1000d     ;(didn't converge)
    #---------------------------------------------------
    fp = self.fp
    f = float64(0.41) * r
    #** f  = 0.5d * r    ;(converged for complex rain, but 0.4 and 0.6 didn't)
    #**                  ;(for simple rain, jumped to f < 0)
    #-----------------
    CONVERGED = False
    tol = float32(1E-8)
    n_tries = int16(0)
    n_max = int16(20)
    while (n_tries < n_max) and not(CONVERGED):
    #----------------------------------------
    #We're using eqn. (6.27) in Smith's book
    #for t(f) and solving it for f(t_end).
    #----------------------------------------
        term1 = (t_end - self.tp) * dK * (float64(1) - gam) / (G * dq)
        term2 = log(float64(1) + (gam * dK) / (f - Ks)) / gam
        term3 = log((f - Ki) / (f - Ks))
        term4 = log(float64(1) + (gam * dK) / (fp - Ks)) / gam
        term5 = log((fp - Ki) / (fp - Ks))
        h = term1 - term2 + term3 + term4 - term5   #(eq 0)
        #---------------------------------------------
        termA = dK / (f - Ks)
        termB = float64(1) / ((f - Ks) + (gam * dK))
        termC = float64(1) / (f - Ki)
        dh_df = termA * (termB - termC)
        del_f = -(h / dh_df)
        f = f + del_f
        CONVERGED = (absolute(del_f) <= tol)
        n_tries = n_tries + 1
    
    #--------------------------------------
    #Did Newton-Raphson fail to converge ?
    #--------------------------------------
    if not(CONVERGED):    
        TF_Print('********************************************')
        TF_Print('  ERROR:  Failure to converge.')
        TF_Print('  Max number of iterations exceeded')
        TF_Print('  while computing infiltration rate.')
        TF_Print('********************************************')
        TF_Print(' ')
        return float64(0)
    
    #----------------------------------------
    #Return infiltration rate at time, t_end
    #----------------------------------------
    self.IN = minimum(f, r)
    ## return minimum(f, r)
    
#  Smith_Parlange_Infil_Rate_1D
#-----------------------------------------------------------------------
def Smith_Parlange_Infil_Rate_3D(self, r, r_last, n):

    Ks  = self.Ks[0]   # (synonyms)
    Ki  = self.Ki[0]
    qs  = self.qs[0]
    qi  = self.qi[0]
    G   = self.G[0]
    gam = self.gam[0]
    
    #----------------------------------
    #Time at start and end of timestep
    #----------------------------------
    t_start = n * self.dt
    t_end = (n + 1) * self.dt
    
    #-----------------------
    #Other common variables
    #-----------------------
    dq  = (qs - qi)
    dK  = (Ks - Ki)
    fac = (G * dq / gam)
    
    #-----------------------------------------------------
    #NB!  wf is where ponding *could* occur during this
    #time step.  Ponding can't occur where (r le *self.Ks)     *** CHECK ***
    #and Ip2 is undefined in this case.
    #-----------------------------------------------------
    wf = where(logical_and((self.tp == -float64(1)), (r > Ks)))
    n_flux = size(wf[0])
    if (n_flux > 0):    
        #--------------------------------------------
        #Compute test values of Ip  (Smith eqn 5.47)
        #--------------------------------------------
        #If (r_last lt Ks) then we'll have Ip1 < 0,
        #so ponding at end of interval can't occur.
        #But it seems it should be able to if r is
        
        #big enough or time step is long enough.
        #So set Ip1 large in this case ??
        #--------------------------------------------
        if (r_last <= Ks):    
            Ip1 = float64(999999)
        else:    
            Ip1 = fac * log(float64(1) + (gam * dK / (r_last - Ks)))
        Ip2 = fac * log(float64(1) + (gam * dK / (r - Ks)))
        dIp = (Ip2 - self.I)
        
        #--------------------------------------
        #Does ponding occur for any grid cells
        #within this timestep ?
        #--------------------------------------
        dIp = (Ip2[wf] - (self.I)[wf])
        wm = where(logical_and((dIp > 0), (dIp < (r[wf] * self.dt))))
        n_mid = size(wm[0])  #******
        if (n_mid != 0):    
            #--------------------
            #Note that: Ip = Ip2
            #--------------------
            self.tp[wf[wm]] = t_start + (dIp[wm] / r[wf[wm]])
            self.fp[wf[wm]] = r[wf[wm]]
        
        #--------------------------------------
        #Does ponding occur for any grid cells
        #at the end of this timestep ?
        #--------------------------------------
        we = where(logical_and((dIp < 0), (self.I < Ip1)))
        n_end = size(we[0])
        if (n_end != 0):    
            #----------------------
            #Note that: Ip = *self.I
            #----------------------
            self.tp[wf[we]] = t_end
            fp_array = ((gam * dK) / (exp(self.I / fac) - float64(1))) + Ks  #***
            self.fp[wf[we]] = fp_array[wf[we]]
        
    
    #------------------------------------------
    #Initialize the grid of infiltration rates
    #------------------------------------------
    ss = idl_func.size(self.tp, dimensions=True)
    nx = ss[0]
    ny = ss[1]
    f = zeros([ny, nx], dtype='Float64')
    
    #---------------------------------------------------
    #For grid cells where ponding has not yet occurred,
    #all "rain" should infiltrate and we set f=r.
    #This will include cases where r < Ks.
    #-------------------------------------------------------
    #Grid cells that *have* ponded but that now have r < Ks        **********
    #are special.  They should also be assigned f = r,
    #since we always have fc > Ks, but are not part of wf.
    #This case is handled at the end when we return (f < R)
    #(the lesser of f and R), since r < Ks < fc.
    #(self.I will be incremented by (f * dt) in caller.)
    #-------------------------------------------------------
    #wp is where ponding has occurred, just now or earlier
    #wf is where ponding has not occurred.
    #-------------------------------------------------------
    wp = where(self.tp != -float64(1))
    n_ponded = size(wp[0])
    wf = where(self.tp == -float64(1))
    n_flux = size(wf[0])
    if (n_flux != 0):    
        f[wf] = r[wf]
    if (n_ponded == 0):    
        return f
    
    #---------------------------------------------------
    #For grid cells where ponding has occurred, we must
    #compute the infiltration rate via grid-based
    #Newton-Raphson iteration using fp and tp arrays.
    #---------------------------------------------------
    #First, we must initialize fp for those grid cells.
    #---------------------------------------------------
    f[wp] = float64(0.41) * r[wp]
    #** f[wp] = 0.5d * r[wp]
    #-----------------------
    CONVERGED = False             #(all grid cells)
    tol = float32(1E-8)            #(tolerance)
    n_tries = int16(0)
    n_max = int16(20)
    while (n_tries < n_max) and not(CONVERGED):
    #----------------------------------------
    #We're using eqn. (6.27) in Smith's book
    #for t(f) and solving it for f(t_end).
    #----------------------------------------
    #Some vars here may be scalars and others
    #may be 2D arrays, but h must be 2D ??
    #We compute h as a 2D array here because        ;**********
    #of this issue, which is not optimal.
    #-----------------------------------------
        term1 = (t_end - self.tp) * dK * (float64(1) - gam) / (G * dq)
        term2 = log(float64(1) + (gam * dK) / (f - Ks)) / gam
        term3 = log((f - Ki) / (f - Ks))
        term4 = log(float64(1) + (gam * dK) / (fp - Ks)) / gam
        term5 = log((fp - Ki) / (fp - Ks))
        h = term1 - term2 + term3 + term4 - term5   #(eq 0)
        #---------------------------------------------
        termA = dK / (f - Ks)
        termB = float64(1) / ((f - Ks) + (gam * dK))
        termC = float64(1) / (f - Ki)
        dh_df = termA * (termB - termC)
        #---------------------------------------------
        del_f = -(h / dh_df)
        f[wp] = f[wp] + del_f[wp]
        CONVERGED = (absolute(del_f[wp]) <= tol)
        n_tries = n_tries + 1
    
    #--------------------------------------
    #Did Newton-Raphson fail to converge ?
    #--------------------------------------
    if not(CONVERGED):    
        TF_Print('********************************************')
        TF_Print('  ERROR:  Failure to converge.')
        TF_Print('  Max number of iterations exceeded')
        TF_Print('  while computing infiltration rate.')
        TF_Print('********************************************')
        TF_Print(' ')
        return float64(0)
    
    #-----------------------------
    #Is r less than Ks anywhere ?
    #If so, set f = r there.
    #-----------------------------
    #***  Check_Low_Rainrate, self, f, r   ;(NO LONGER NEEDED, SEE ABOVE)
    
    #-----------------------------------------
    #Return infiltration rates at time, t_end
    #-----------------------------------------
    self.IN = minimum(f, r)
    ## return minimum(f, r)
    
#   Smith_Parlange_Infil_Rate_3D
#-----------------------------------------------------------------------
def Smith_Parlange_Infil_Rate_v2(self, r, r_last, n):

    #------------------------------------------------------------
    #Notes:  This was written on 5/11/06 using information
    #        from Smith, R.E. (2002) Infiltration Theory for
    #        Hydrologic Applications, Water Resources Monograph
    #        15, AGU, pages 114-116.

    #        IN  = infiltration rate [m/s]
    #        Ks  = saturated hydraulic conductivity [m/s]
    #        Ki  = initial hydraulic conductivity [m/s]
    #        qs  = soil moisture content [dimless]
    #        qi  = soil initial moisture content [dimless]
    #         G  = capillary length scale [m]
    #         I  = cum. infiltration depth (since reset) [m]
    #         P  = precipitation rate [m/s]
    #        SM  = snowmelt rate [m/s]
    #         n  = time step (for computing t_start & t_end)

    #        Note that the infiltration rate has a max possible
    #        value of (P + SM) and asymptotes to Ks as the
    #        total infiltrated depth increases.  Need to reset
    #        this depth between "events", but haven't decided
    #        how to do this yet.

    #        Total infiltrated depth is incremented in the
    #        calling function, called Infiltration.
    #------------------------------------------------------------
    SINGLE_PROFILE = self.all_scalars  #(3/19/07)
    
    if (SINGLE_PROFILE):    
        Smith_Parlange_Infil_Rate_1D(self, r, r_last, n)
    else:    
        Smith_Parlange_Infil_Rate_3D(self, r, r_last, n)
    
#  Smith_Parlange_Infil_Rate_v2
#-----------------------------------------------------------------------
