
## Copyright (c) 2001-2010, Scott D. Peckham
## January 2009  (converted from IDL)
## August 2009
## May 2010 (changes to initialize() and read_cfg_file()
## June 2010 (update_net_shortwave_radiation, etc.)  

#-----------------------------------------------------------------------
#  NOTES:  This file defines a "base class" for meteorology
#          components as well as any functions used by most or
#          all meteorology methods.  The methods of this class
#          should be over-ridden as necessary for different
#          methods of modeling meteorology.
#-----------------------------------------------------------------------
#
#  unit_test()
#
#  class met_component     (inherits from CSDMS_base.py)
#
#      get_attribute()       # (10/26/11)
#
#      get_input_items()     # (not ready yet)
#      get_output_items()
#      set_constants()
#      ---------------------
#      initialize()
#      update()
#      finalize()
#      ----------------------------
#      set_computed_input_vars()
#      initialize_computed_vars()
#      ----------------------------
#      get_cca_port_info()           # (5/11/10)
#      embed_child_components()
#      add_child_ports()
#      initialize_ports()
#      get_cca_ports()               # (in CSDMS_base.py)
#      release_cca_ports()           # (in CSDMS_base.py)
#      ------------------------------------
#      update_P_integral()
#      update_P_max()
#      update_snow_vars()
#      ------------------------------------
#      update_richardson_number()
#      update_bulk_exchange_coeff()
#      update_sensible_heat_flux()
#      update_saturation_vapor_pressure()
#      update_vapor_pressure()
#      update_dew_point()                    # (7/6/10)
#      update_precipitable_water_content()   # (7/6/10)
#      ------------------------------------
#      update_latent_heat_flux()
#      update_conduction_heat_flux()
#      update_advection_heat_flux()
#      ------------------------------------
#      update_julian_day()                   # (7/1/10)
#      update_net_shortwave_radiation()      # (7/1/10)
#      update_em_air()                       # (7/1/10)
#      update_net_longwave_radiation()       # (7/1/10)
#      update_net_energy_flux()              # ("Q_sum")
#      ------------------------------------
#      open_input_files()
#      read_input_files()
#      close_input_files()
#      ------------------------------------
#      update_outfile_names()
#      open_output_files()
#      write_output_files()
#      close_output_files()
#      save_grids()
#      save_pixel_values()
#
#  Functions:
#      compare_em_air_methods()
#
#-----------------------------------------------------------------------

from numpy import *
import numpy

import os

import CSDMS_base
import model_input
import model_output
import solar_funcs as solar
import rtg_files
import tf_utils

## from model_output import *

#-----------------------------------------------------------------------
def unit_test():

    c = met_component()
    c.CCA   = False
    c.DEBUG = True
    
    #-----------------------------------------
    # This function adjusts for the platform
    # and can be changed in "tf_utils.py".
    #-----------------------------------------
    tf_utils.TF_Set_Test_Info( c ) 

    #-------------------------------------------
    # Call initialize() and call update() once
    #-------------------------------------------
##    print 'STATUS =', c.get_status()
##    c.initialize( mode="driver" )
##    print 'STATUS =', c.get_status()
##    time_sec = float64(0)
##    c.update(time_sec)
##    print 'STATUS =', c.get_status()

    c.run_model( cfg_directory = c.cfg_directory,
                 cfg_prefix = c.cfg_prefix )
  
#   unit_test()
#-----------------------------------------------------------------------
class met_component(CSDMS_base.CSDMS_component):

        #-------------------------------------------------------------
        # Notes: **************************************************
        #        Do we ever need to distinguish between a surface
        #        temperature and snow temperature (in the snow) ?
        #        Recall that a separate T_soil_x variable is used
        #        to compute Qc.
        #        **************************************************

        #        Vapor_Pressure function is defined in Qnet_file.py
        #        and accepts temperature argument, in Celsius, and
        #        relative humidity.  It returns vapor pressure in
        #        units of kPa.

        #        Cp_snow is from NCAR CSM Flux Coupler web page

        #        uz = wind speed, z = reference height above ground

        #        rho_H2O is currently not adjustable with GUI.
        #------------------------------------------------------------

    #-------------------------------------------------------------------
    def get_attribute(self, att_name):

        #----------------
        # New. 10/26/11
        #----------------
        map = {'comp_name':          'Meteorology',
               'version':            '3.1',
               'model_name':         'Meteorology',
               'model_family':       'TopoFlow',
               'cfg_template_file':  'Meteorology.cfg.in',
               'cfg_extension':      '_meteorology.cfg',
               'cmt_var_prefix':     '/Meteorology/Input/Var/',
               'gui_xml_file':       '/home/csdms/cca/topoflow/3.1/src/share/cmt/gui/Meteorology.xml',
               'dialog_title':       'Meteorology: Method 1 Parameters',
               'time_step_type':     'fixed',
               'time_units':         'seconds',
               'mesh_type':          'none',
               'author_name':        'Scott Peckham'}

        try:
            return map[ att_name.lower() ]
        except:
            print '###################################################'
            print ' ERROR: Could not find attribute: ' + att_name
            print '###################################################'
            print ' '

    #   get_attribute() 
    #-------------------------------------------------------------------
    def set_constants(self):

        #---------------------------------
        # Define some physical constants
        #---------------------------------
        self.g        = float64(9.81)    # [m/s^2, gravity]
        self.kappa    = float64(0.408)   # [none, von Karman]
        self.rho_H2O  = float64(1000)
        self.rho_air  = float64(1.2614)
        self.Cp_air   = float64(1005.7)
        self.Lv       = float64(2500000) # [Joules/kg, Latent heat of vaporiz.)
        self.Lf       = float64(334000)  # [J/kg = W*s/kg, Latent heat of fusion)
        self.sigma    = float64(5.67E-8) # [W/(m^2 K^4)]  (Stefan-Boltzman constant)
        self.C_to_K   = float64(273.15)  # (add to convert [deg_C] to [deg_K])

        self.twopi         = float64(2) * numpy.pi
        self.one_seventh   = float64(1) / 7
        self.hours_per_day = float64(24)
        self.secs_per_day  = float64(3600) * self.hours_per_day
 
        #----------------------------------------
        # Constants related to precip (9/24/09)
        #----------------------------------------
        self.mmph_to_mps = (float64(1) / float64(3600000))
        self.mps_to_mmph = float64(3600000)
        self.forever     = float64(999999999)  # [minutes]
        
        #------------------------------------------------
        # Only needed for method 1, where all rates and
        # durations are read as 1D arrays from GUI.
        # Method 1 may be removed in a future version.
        #------------------------------------------------
##        self.method1_rates     = None
##        self.method1_durations = None
##        self.method1_n_rates   = 0
        
    #   set_constants()             
    #-------------------------------------------------------------------
    def initialize(self, cfg_prefix=None, mode="nondriver",
                   SILENT=False):

        if not(SILENT):
            print ' '
            print 'Meteorology component: Initializing...'
            
        self.status     = 'initializing'  # (OpenMI 2.0 convention)
        self.mode       = mode
        self.cfg_prefix = cfg_prefix
        
        #-----------------------------------------------
        # Load component parameters from a config file
        #-----------------------------------------------
        self.set_constants()
        self.initialize_config_vars()
        ## print '    Calling read_grid_info()...'
        self.read_grid_info()
        ## print '    Calling initialize_basin_vars()...'
        self.initialize_basin_vars()  # (5/14/10)
        #----------------------------------------------------
        # NB! This read_input_files() uses self.time_index.
        #     Also needs to be before "Disabled" test.
        #----------------------------------------------------
        ## print '    Calling initialize_time_vars()...'
        self.initialize_time_vars()
        
        #------------------------------------------------------
        # NB! "Sample steps" must be defined before we return
        #     Check all other process modules.
        #------------------------------------------------------
        if (self.comp_status == 'Disabled'):
            if not(SILENT):
                print 'Meteorology component: Disabled.'
            self.P        = float64(0)
            self.e_air    = float64(0)
            self.e_surf   = float64(0)
            self.em_air   = float64(0)
            self.Qn_SW    = float64(0)
            self.Qn_LW    = float64(0)
            self.Q_sum    = float64(0)
            self.DONE     = True
            self.status   = 'initialized'
            return

        #-----------------------------------------------
        # Read from files as needed to initialize vars 
        #-----------------------------------------------
        self.open_input_files()
        self.read_input_files()  # (initializes P)
        
        self.initialize_required_components(mode)
        ## self.check_input_types()  # (not needed so far)
        
        #-----------------------
        # Initialize variables
        #-----------------------
        ## print '    Calling initialize_computed_vars()...'
        self.initialize_computed_vars() # (after read_input_files)
        
        if not(self.PRECIP_ONLY):
            self.open_output_files() 
        self.status = 'initialized'  # (OpenMI 2.0 convention)
        
    #   initialize()
    #-------------------------------------------------------------------
    def update(self, time_seconds=None):

        #----------------------------------------------------------
        # Note: The read_input_files() method is first called by
        #       the initialize() method.  Then, the update()
        #       method is called one or more times, and it calls
        #       other update_*() methods to compute additional
        #       variables using input data that was last read.
        #       Based on this pattern, read_input_files() should
        #       be called at end of update() method as done here.
        #       If the input files don't contain any additional
        #       data, the last data read persists by default.
        #----------------------------------------------------------
        if (self.comp_status == 'Disabled'): return
        self.status = 'updating'  # (OpenMI 2.0 convention)

        #-------------------------------------------
        # Update computed values related to precip
        #-------------------------------------------
        self.update_P_integral()
        self.update_P_max()
        self.update_snow_vars()   # (snow depth due to snowfall)

        #-------------------------
        # Update computed values
        #-------------------------
        if not(self.PRECIP_ONLY):
            self.update_richardson_number()
            self.update_bulk_exchange_coeff()
            self.update_sensible_heat_flux()
            self.update_saturation_vapor_pressure(MBAR=True)
            self.update_saturation_vapor_pressure(MBAR=True, SURFACE=True)  ##########
            self.update_vapor_pressure(MBAR=True)
            self.update_dew_point() ###
            self.update_precipitable_water_content()  ###
            self.update_vapor_pressure(MBAR=True, SURFACE=True)   ########
            self.update_latent_heat_flux()      # (uses e_air and e_surf)
            self.update_conduction_heat_flux()
            self.update_advection_heat_flux()
            self.update_julian_day()
            self.update_net_shortwave_radiation()
            self.update_em_air()
            self.update_net_longwave_radiation()
            self.update_net_energy_flux()  # (at the end)
            
        #----------------------------------------
        # Read next met vars from input files ?
        #-------------------------------------------
        # Note that read_input_files() is called
        # by initialize() and these values must be
        # used for "update" calls before reading
        # new ones.
        #-------------------------------------------
        if (self.time_index > 0):
            self.read_input_files()
      
        #----------------------------------------------
        # Write user-specified data to output files ?
        #----------------------------------------------
        if not(self.PRECIP_ONLY):
            self.write_output_files(time_seconds)

        #-----------------------------
        # Update internal clock
        # after write_output_files()
        #-----------------------------
        self.update_time()
        self.status = 'updated'  # (OpenMI)
        
    #   update()
    #-------------------------------------------------------------------
    def finalize(self):

        self.status = 'finalizing'  # (OpenMI)
        if (self.comp_status == 'Enabled'):
            self.close_input_files()   ##  TopoFlow input "data streams"
            if not(self.PRECIP_ONLY):
                self.close_output_files()
        self.status = 'finalized'  # (OpenMI)

        self.print_final_report(comp_name='Meteorology component')
        
        #---------------------------
        # Release all of the ports
        #----------------------------------------
        # Make this call in "finalize()" method
        # of the component's CCA Imple file
        #----------------------------------------
        # self.release_cca_ports( port_names, d_services )
        
    #   finalize()
    #-------------------------------------------------------------------
    def set_computed_input_vars(self):

        #-----------------------------------------------
        # Convert precip rate units from mm/h to m/s ?
        #-----------------------------------------------
        # NB! read_input_files() does this for files.
        #-----------------------------------------------
        if (self.P_type == 'Scalar'):
##            print '######## self.P_type  =', self.P_type
##            print '######## type(self.P) =', type(self.P)
##            print '######## self.P       =', self.P
##            print '######## Converting scalar P from MMPH to MPS.'
            self.P = self.P * self.mmph_to_mps
            print 'Scalar rainrate set to:', self.P, ' [mmph]'
            
        #---------------------------------
        # Process the PRECIP_ONLY toggle
        #---------------------------------
        if not(hasattr(self, 'PRECIP_ONLY')):
            self.PRECIP_ONLY = False
        elif (self.PRECIP_ONLY.lower() == 'yes'):
            self.PRECIP_ONLY = True
        else:
            self.PRECIP_ONLY = False

        #---------------------------------------
        # Print info message about PRECIP_ONLY
        #---------------------------------------
        if (self.PRECIP_ONLY):
            print '-----------------------------------------'
            print ' NOTE: Since PRECIP_ONLY = True, output'
            print '       variables will not be computed'
            print '       or saved to files.'
            print '-----------------------------------------'
            print' '
            
        #----------------------------------------------------
        # Toggle to use SATTERLUND or BRUTSAERT methods
        # for computing e_air and em_air. (Not in GUI yet.)
        #----------------------------------------------------
        if not(hasattr(self, 'SATTERLUND')):
            self.SATTERLUND = False

        #---------------------------------------------
        # Convert GMT_offset from string to int
        # because GUI can't use ints in droplist yet
        #---------------------------------------------
        self.GMT_offset = numpy.int16( self.GMT_offset )

        #------------------------------------------------
        # Convert start_month from string to integer
        # January should be 1.  See solar.Julian_Day().
        #------------------------------------------------
        month_list = ['January', 'February', 'March', 'April',
                      'May', 'June', 'July', 'August', 'September',
                      'October', 'November', 'December']
        self.start_month = month_list.index( self.start_month ) + 1
                 
        #-------------------------------
        # Initialize some more toggles
        #-------------------------------                   
        if not(hasattr(self, 'SAVE_QSW_GRIDS')):
            self.SAVE_QSW_GRIDS = False
        if not(hasattr(self, 'SAVE_QLW_GRIDS')):
            self.SAVE_QLW_GRIDS = False
        #-------------------------------------------
        if not(hasattr(self, 'SAVE_QSW_PIXELS')):
            self.SAVE_QSW_PIXELS = False
        if not(hasattr(self, 'SAVE_QLW_PIXELS')):
            self.SAVE_QLW_PIXELS = False
   
        #---------------------------------------------------------
        # Make sure that all "save_dts" are larger or equal to
        # the specified process dt.  There is no point in saving
        # results more often than they change.
        # Issue a message to this effect if any are smaller ??
        #---------------------------------------------------------
        self.save_grid_dt   = maximum(self.save_grid_dt,    self.dt)
        self.save_pixels_dt = maximum(self.save_pixels_dt,  self.dt)
        
    #   set_computed_input_vars()
    #-------------------------------------------------------------------
    def initialize_computed_vars(self):

        #------------------------------------------------------
        # Note: Some of these require "self.rti", which is
        #       only stored by read_grid_info() after the
        #       set_computed_input_vars() function is called.
        #       So these parts can't go there.
        #------------------------------------------------------
        
        #---------------------------------------
        # Add self.in_directory to:
        #   slope_grid_file & aspect_grid_file
        #---------------------------------------
        self.slope_grid_file  = (self.in_directory + self.slope_grid_file) 
        self.aspect_grid_file = (self.in_directory + self.aspect_grid_file)
        
        #-------------------------------------------------
        # Read slope grid & convert to slope angle, beta
        # NB!  RT slope grids have NaNs on edges.
        #-------------------------------------------------
        slopes = rtg_files.read_grid( self.slope_grid_file, self.rti,
                                      RTG_type='FLOAT' )
        beta   = numpy.arctan( slopes )
        beta   = (self.twopi + beta) % self.twopi
        #---------------------------------------------
        w_nan = where( logical_not(isfinite(beta)) )
        n_nan = numpy.size(w_nan[0])
        if (n_nan != 0):    
            beta[ w_nan ] = numpy.float64(0)
        #------------------------------------------------------------------
        w_bad = where( logical_or( (beta < 0), (beta > numpy.pi / 2) ) )
        n_bad = size(w_bad[0])
        if (n_bad != 0):    
            msg = array(['ERROR:  Some slope angles are out of range.', ' '])
            for line in msg:
                print line
            ## result = GUI_Message(msg, INFO=True, TITLE='ERROR MESSAGE')
            return
        self.beta = beta  ######
        
        #------------------------------------------------------
        # Read aspect grid.  Alpha must be CW from north.
        # NB!  RT aspect grids have NaNs on edges.
        #---------------------------------------------------------
        # RT files ending in "_mf-angle.rtg" and "fd-aspect.rtg"
        # contain aspect values.  The former are in [0, 2 Pi]
        # while the latter are in [-Pi, Pi] and both measure
        # CCW from due east.
        #---------------------------------------------------------
        aspects = rtg_files.read_grid( self.aspect_grid_file, self.rti,
                                       RTG_type='FLOAT' )
        alpha   = (numpy.pi / 2) - aspects
        alpha   = (self.twopi + alpha) % self.twopi
        #-----------------------------------------------
        w_nan = where( logical_not(isfinite(alpha)) )
        n_nan = numpy.size( w_nan[0] )
        if (n_nan != 0):    
            alpha[ w_nan ] = numpy.float64(0)
        self.alpha = alpha  ######

        #---------------------------        
        # Create lon and lat grids
        #---------------------------
        if (self.rti.pixel_geom == 0):
            self.lon_deg = solar.Longitude_Grid( self.rti )
            self.lat_deg = solar.Latitude_Grid( self.rti )

##            print 'Lon grid ='
##            print self.lon_deg
##            print 'Lat grid ='
##            print self.lat_deg
            
            #-----------------------------
            # Write grids to RTG files ?
            #-----------------------------
##            lon_file = (self.out_directory + self.site_prefix + '_lons.bin')
##            rtg_files.write_grid( self.lon_deg, lon_file, self.rti )
##            lat_file = (self.out_directory + self.site_prefix + '_lats.bin')
##            rtg_files.write_grid( self.lat_deg, lat_file, self.rti )
        else:    
            print 'SORRY: Cannot yet create lon and lat grids for'
            print '       this DEM because it uses UTM coordinates.'
            print '       Will use lat/lon for Denver, Colorado.'
            print ' '
            #--------------------------------------------
            # For now, use scalar values for Denver, CO
            #--------------------------------------------
            self.lon_deg = numpy.float64( -104.9841667 )
            self.lat_deg = numpy.float64( 39.7391667 )
            ## return

        #-------------------------------------------------
        # Initialize max precip rate with the first rate
        #------------------------------------------------
        # Note: Need this here because rate may be
        #       zero at the end of update_precip_rate()
        #------------------------------------------------
        self.P_max = self.P.max()    # (after read_input_files)
        self.vol_P = float64(0)      # (for mass balance)
        self.Q_sum = float64(0)      # (for energy_balance snow + ET)

        #------------------------------------
        # Initialize the decimal Julian day
        #------------------------------------
        self.julian_day = solar.Julian_Day( self.start_month,
                                            self.start_day,
                                            self.start_hour )
        ## print '    julian_day =', self.julian_day
        
    #   initialize_computed_vars()
    #-------------------------------------------------------------------
    def get_cca_port_info(self):

        self.cca_port_names = ['snow']
        self.cca_port_short = ['sp']
        self.cca_port_type  = "IRFPort"
        self.cca_project    = "edu.csdms.models"

    #   get_cca_port_info()        
    #-------------------------------------------------------------------
    def embed_child_components(self):

        #------------------------------------------------
        # Note: Don't call this if (self.CCA == True).
        #------------------------------------------------
        # Instantiate and embed "process components"
        # in the place of the CCA ports.
        #------------------------------------------------
        # But how do we choose a given process method
        # such as "kinematic wave" ??
        #------------------------------------------------
        import snow_degree_day
        
        self.sp = snow_degree_day.snow_component()
        ###  self.pp = precip_base.precip_component()
        
    #   embed_child_components()
    #-------------------------------------------------------------------
    def add_child_ports(self):

        self.add_child_port('sp', 'mp', SELF=True)
        
    #   add_child_ports()        
    #-------------------------------------------------------------------
    def initialize_ports(self):
        
        #-----------------------------------------------
        # Initialize the process objects/components
        # This is also where output files are opened.
        #-----------------------------------------------
        if (self.sp.get_status() != 'initialized'):         # snow vars         
            self.sp.initialize( cfg_prefix=self.cfg_prefix )

##        if (self.pp.get_status() != 'initialized'):        # precip vars         
##            self.pp.initialize( cfg_prefix=self.cfg_prefix )
            
    #   initialize_ports()
    #-------------------------------------------------------------------
    def update_P_integral(self):

        #---------------------------------------------------
        # Notes: This can be used for mass balance checks,
        #        such as now done by update_mass_totals()
        #        in topoflow.py.  The "dt" here should be
        #        TopoFlow's "main dt" vs. the process dt.
        
        #        dV[i] = P[i] * da[i] * dt, dV = sum(dV[i])
        #---------------------------------------------------
        if (self.DEBUG):
            print 'Calling update_P_integral()...'
            
        #------------------------------------------------
        # Update mass total for P, sum over all pixels
        #------------------------------------------------   
        volume = double(self.P * self.da * self.dt)  # [m^3]
        if (size(volume) == 1):
            self.vol_P += (volume * self.rti.n_pixels)
        else:
            self.vol_P += sum(volume)
        
    #   update_P_integral()
    #-------------------------------------------------------------------
    def update_P_max(self):

        if (self.DEBUG):
            print 'Calling update_P_max()...'
            
        #-----------------------------------------
        # Save the maximum precip. rate in [m/s]
        #-----------------------------------------
        self.P_max = maximum(self.P_max, self.P.max())

    #   update_P_max()   
    #-------------------------------------------------------------------
    def update_snow_vars(self):

        if (self.DEBUG):
            print 'Calling update_snow_vars()...'
            
        #---------------------------------------------
        # If P or T_air is a grid, then we must have
        # h_swe and h_snow be grids.  This is set
        # up at start of Route_Flow.
        #---------------------------------------------
        h_swe    = self.get_port_data('h_swe',  self.sp)
        h_snow   = self.get_port_data('h_snow', self.sp)
        ## print 'MADE IT PAST get_port_data() calls.'

        #--------------------------------------------------        
        rho_H2O  = self.sp.get_scalar_double('rho_H2O')
        rho_snow = self.sp.get_scalar_double('rho_snow')
        ### main_dt  = self.cp.get_scalar_double('dt')
        ## print 'MADE IT PAST get_scalar_double() calls.'
        
        #----------------------------------------------
        # Update snow water equivalent and snow depth
        #----------------------------------------------
        # (3/14/07) New method that works regardless
        # of whether P and T are scalars or grids.
        #---------------------------------------------
        # (8/20/09) Precip process now has constant
        # timestep, dt, which should be used here.
        #---------------------------------------------        
        dh_swe  = (self.P * (self.T_air <= 0)) * self.dt
        h_swe  += dh_swe
        ratio   = (rho_H2O / rho_snow)
        h_snow  = h_swe * ratio
        ## print 'MADE IT PAST var updates.'
        
        #----------------------------------------------------
        # Use "set_values" call to update h_snow and h_swe?
        #----------------------------------------------------
##        print 'shape(h_snow) =', shape(h_snow)
##        print 'shape(h_swe)  =', shape(h_swe)
##        print 'shape(P)      =', shape(self.P)
##        print 'shape(T_air)  =', shape(self.T_air)
##        print 'T_air         =', self.T_air
        
        self.set_port_data('h_snow', h_snow, self.sp)
        self.set_port_data('h_swe',  h_swe,  self.sp)
        ## print 'MADE IT PAST set_port_data() calls.'
        
        #---------------------------------------------
        # Where precip falls as snow and gets saved
        # to SWE, P must be set to zero so it can't
        # go on to produce runoff.
        #-----------------------------------------------
        # A special situation occurs if P is a scalar
        # and T is a grid.  P must then be converted
        # to a grid in order to be able to reflect the
        # fact that water was stored as snow in some
        # pixels and not at others.  This happens
        # automatically as written here. (3/14/07)
        # If P and T are both scalars, P stays scalar.
        ################################################
        # NB!  Should still work in Python, because:
        #   (1.5 * True = 1.5, 1.5 * False = 0)
        ################################################
        # NB!  This could be slowing things down a lot
        #      because it happens every timestep, not
        #      just when precip rate is updated.  Is
        #      there some way to test out of this ??
        ################################################
        ################################################
        self.P = (self.P * (dh_swe == 0))

        #--------------
        # For testing
        #--------------
        # nT = size(mv.T_air)
        # nP = size(self.P)
        # if (nT == 1):
        #     tstr = str(mv.T_air)
        #     print '   T_air = ' + tstr
        # if (nP == 1):
        #     pstr = str(self.P)
        #     print '   P     = ' + pstr
        
        #-------------------------------------
        # Message about rainfall or snowfall
        #-------------------------------------
        # Note that rain and snow may fall
        # simultaneously at different pixels
        #-------------------------------------
        # Reporting this slows model too much
        # but can be used for testing
        #-------------------------------------
        # w1 = where(dh_swe > 0)
        # n1 = size(w1[0])
        # if (n1 > 0): print '   Snow is falling...'
        #--------------------------------------------------
        # dh_rain = (P * (mv.T_air > 0))
        # w2 = where(dh_rain > 0)
        # n2 = size(w2[0])
        # if (n2 > 0): print '   Rain is falling...'

    #   update_snow_vars() 
    #-------------------------------------------------------------------
    def update_richardson_number(self):

        if (self.DEBUG):
            print 'Calling update_richardson_number()...'
            
        #---------------------------------------------------------------
        # Notes: Other definitions are possible, such as the one given
        #        by Dingman (2002, p. 599).  However, this one is the
        #        one given by Zhang et al. (2000) and is meant for use
        #        with the stability criterion also given there.
        #---------------------------------------------------------------
        top     = self.g * self.z * (self.T_air - self.T_surf)
        bot     = (self.uz)**2.0 * (self.T_air + float64(273.15))
        self.Ri = (top / bot)

    #   update_richardson_number()
    #-------------------------------------------------------------------        
    def update_bulk_exchange_coeff(self):

        if (self.DEBUG):
            print 'Calling update_bulk_exchange_coeff()...'

        #----------------------------------------------------------------
        # Notes: Dn       = bulk exchange coeff for the conditions of
        #                   neutral atmospheric stability [m/s]
        #        Dh       = bulk exchange coeff for heat  [m/s]
        #        De       = bulk exchange coeff for vapor [m/s]
        #        h_snow   = snow depth [m]
        #        z0_air   = surface roughness length scale [m]
        #                   (includes vegetation not covered by snow)
        #        z        = height that has wind speed uz [m]
        #        uz       = wind speed at height z [m/s]
        #        kappa    = 0.408 = von Karman's constant [unitless]
        #        RI       = Richardson's number (see function)
        #----------------------------------------------------------------
        h_snow = self.get_port_data('h_snow', self.sp)
        
        #---------------------------------------------------
        # Compute bulk exchange coeffs (neutral stability)
        # using the logarithm "law of the wall".
        #---------------------------------------------------
        arg = self.kappa / log((self.z - h_snow) / self.z0_air)
        Dn  = self.uz * (arg)**2.0
        
        #-----------------------------------------------
        # NB! Dn could be a scalar or a grid, so this
        #     must be written to handle both cases.
        #     Note that WHERE can be used on a scalar:
        
        #     IDL> a = 1
        #     IDL> print, size(a)
        #     IDL> w = where(a ge 1, nw)
        #     IDL> print, nw
        #     IDL> a[w] = 2
        #     IDL> print, a
        #     IDL> print, size(a)
        #-----------------------------------------------

        ###########################################################
        #  NB!  If T_air and T_surf are both scalars, then next
        #       few lines won't work because we can't index the
        #       resulting empty "w" (even if T_air == T_surf).
        ###########################################################
##        w  = where(self.T_air != self.T_surf)
##        nw = size(w[0])
##        ## nw = size(w,0)  # (doesn't work if 2 equal scalars)
        #----------------------------------------------------------
        T_AIR_SCALAR  = (rank(self.T_air)  == 0)
        T_SURF_SCALAR = (rank(self.T_surf) == 0)
        if (T_AIR_SCALAR and T_SURF_SCALAR):
            if (self.T_air == self.T_surf):  nw=1
            else: nw=0      
        else:
            w  = where(self.T_air != self.T_surf)
            nw = size(w[0])
        #----------------------------------------------------------            
        if (nw == 0):
            self.Dn = Dn
            self.Dh = Dn
            self.De = Dn
            return
        
        #-------------------------------------
        # One or more pixels are not neutral
        # so make a correction using RI
        #---------------------------------------------
        # NB!  RI could be a grid when Dn is a
        # scalar, and this will change Dn to a grid.
        #---------------------------------------------
        # Ri = Richardson_Number(z, uz, T_air, T_surf)
        #--------------------------------------------
        # Before 12/21/07.  Has bug if RI is a grid
        #--------------------------------------------
        # w_stable = where(*T_air gt *T_surf, n_stable)
        # if (n_stable ne 0) then begin
        #     Dn[w_stable] = Dn[w_stable]/(1d + (10d * RI))
        # endif
        # w_unstable = where(*T_air lt *T_surf, n_unstable)
        # if (n_unstable ne 0) then begin
        #----------------------------------------------
        # Multiplication and substraction vs. opposites
        # for the stable case.  Zhang et al. (2000)
        # Hopefully not just a typo.
        #----------------------------------------------
        #    Dn[w_unstable] = Dn[w_unstable]*(1d - (10d * self.Ri))
        # endif
        
        #-----------------
        # After 12/21/07
        #------------------------------------------------------------
        # If T_air, T_surf or uz is a grid, then Ri will be a grid.
        # This version makes only one call to WHERE, so its faster.
        #------------------------------------------------------------
        # Multiplication and substraction vs. opposites for the
        # stable case (Zhang et al., 2000); hopefully not a typo.
        # It plots as a smooth curve through Ri=0.
        #------------------------------------------------------------
        nD = size(Dn)
        nR = size(self.Ri)
        if (nR > 1):    
            #--------------------------
            # Case where RI is a grid
            #--------------------------
            ws = where(self.Ri > 0)
            ns = size(ws[0])
            wu = where(invert(self.Ri > 0))
            nu = size(wu[0])
            if (nD == 1):    
                #******************************************
                # Convert Dn to a grid here or somewhere
                # Should stop with an error message
                #******************************************
                dum = int16(0)
            if (ns != 0):    
                Dn[ws] = Dn[ws] / (float64(1) + (float64(10) * self.Ri[ws]))
            if (nu != 0):    
                Dn[wu] = Dn[wu] * (float64(1) - (float64(10) * self.Ri[wu]))
        else:    
            #----------------------------
            # Case where Ri is a scalar
            #--------------------------------
            # Works if Dn is grid or scalar
            #--------------------------------
            if (self.Ri > 0):    
                Dn = Dn / (float64(1) + (float64(10) * self.Ri))
            else:    
                Dn = Dn * (float64(1) - (float64(10) * self.Ri))

        #----------------------------------------------------
        # NB! We currently assume that these are all equal.
        #----------------------------------------------------
        self.Dn = Dn
        self.Dh = Dn
        self.De = Dn
        
    #   update_bulk_exchange_coeff()
    #-------------------------------------------------------------------
    def update_sensible_heat_flux(self):

        #--------------------------------------------------------
        # Notes: All the Q's have units of W/m^2 = J/(m^2 s).
        #        Dh is returned by Bulk_Exchange_Coeff function
        #        and is not a pointer.
        #--------------------------------------------------------
        if (self.DEBUG):
            print 'Callilng update_sensible_heat_flux()...'
            
        #---------------------
        # Physical constants
        #---------------------
        # rho_air = 1.225d   ;[kg/m^3, at sea-level]
        # Cp_air  = 1005.7   ;[J/kg/deg_C]
        
        #-----------------------------
        # Compute sensible heat flux
        #-----------------------------
        delta_T = (self.T_air - self.T_surf)
        self.Qh = (self.rho_air * self.Cp_air) * self.Dh * delta_T

    #   update_sensible_heat_flux()
    #-------------------------------------------------------------------
    def update_saturation_vapor_pressure(self, MBAR=False,
                                         SURFACE=False):

        if (self.DEBUG):
            print 'Calling update_saturation_vapor_pressure()...'

        #----------------------------------------------------------------
        #Notes:  Saturation vapor pressure is a function of temperature.
        #        T is temperature in Celsius.  By default, the method
        #        of Brutsaert (1975) is used.  However, the SATTERLUND
        #        keyword is set then the method of Satterlund (1979) is
        #        used.  When plotted, they look almost identical.  See
        #        the Compare_em_air_Method routine in Qnet_file.pro.
        #        Dingman (2002) uses the Brutsaert method.
        #        Liston (1995, EnBal) uses the Satterlund method.

        #        By default, the result is returned with units of kPa.
        #        Set the MBAR keyword for units of millibars.
        #        100 kPa = 1 bar = 1000 mbars
        #                => 1 kPa = 10 mbars
        #----------------------------------------------------------------
        #NB!     Here, 237.3 is correct, and not a misprint of 273.2.
        #        See footnote on p. 586 in Dingman (Appendix D).
        #----------------------------------------------------------------
        if (SURFACE):
##            if (self.T_surf_type in ['Scalar', 'Grid']):
##                return
            T = self.T_surf
        else:
##            if (self.T_air_type in ['Scalar', 'Grid']):
##                return
            T = self.T_air
        
        if not(self.SATTERLUND):    
            #------------------------------
            # Use Brutsaert (1975) method
            #------------------------------
            term1 = (float64(17.3) * T) / (T + float32(237.3))
            e_sat = float64(0.611) * exp(term1)        # [kPa]
        else:    
            #-------------------------------
            # Use Satterlund (1979) method
            #-------------------------------
            e_sat = float64(10) ** (float64(11.4) - (float64(2353) / (T + float32(273.15))))   # [Pa]
            e_sat = (e_sat / float64(1000))   # [kPa]

        #-----------------------------------
        # Convert units from kPa to mbars?
        #-----------------------------------
        if (MBAR):    
            e_sat = (e_sat * numpy.float64(10))   # [mbar]

        if (SURFACE):
            self.e_sat_surf = e_sat
        else:
            self.e_sat_air  = e_sat

    #   update_saturation_vapor_pressure()
    #------------------------------------------------------------------- 
    def update_vapor_pressure(self, MBAR=False,
                              SURFACE=False):

        if (self.DEBUG):
            print 'Calling update_vapor_pressure()...'

        #---------------------------------------------------
        # Notes: T is temperature in Celsius
        #        RH = relative humidity, in [0,1]
        #             by definition, it equals (e / e_sat)
        #        e has units of kPa.
        #---------------------------------------------------
        if (SURFACE):
##            if (self.T_surf_type in ['Scalar', 'Grid']) and \
##               (self.RH_type in ['Scalar', 'Grid']):
##                return
            e_sat = self.e_sat_surf
        else:
##            if (self.T_air_type in ['Scalar', 'Grid']) and \
##               (self.RH_type in ['Scalar', 'Grid']):
##                return
            e_sat = self.e_sat_air
            
        e = (self.RH * e_sat)
        
        #-----------------------------------
        # Convert units from kPa to mbars?
        #-----------------------------------
        if (MBAR):    
            e = (e * numpy.float64(10))   # [mbar]

        if (SURFACE):
            self.e_surf = e
        else:
            self.e_air  = e
 
    #   update_vapor_pressure()
    #-------------------------------------------------------------------
    def update_dew_point(self):

        if (self.DEBUG):
            print 'Calling update_dew_point()...'

        #-----------------------------------------------------------
        # Notes:  The dew point is a temperature in degrees C and
        #         is a function of the vapor pressure, e_air.
        #         Vapor pressure is a function of air temperature,
        #         T_air, and relative humidity, RH.
        #         The formula used here needs e_air in kPa units.
        #-----------------------------------------------------------
        e_air_kPa = self.e_air / numpy.float64(10)   # [kPa]
        log_vp    = numpy.log( e_air_kPa )

        top = log_vp + float64(0.4926)
        bot = float64(0.0708) - (float64(0.00421) * log_vp)

        self.T_dew = (top / bot)    # [degrees C]
    
    #   update_dew_point()
    #-------------------------------------------------------------------
    def update_precipitable_water_content(self):

        if (self.DEBUG):
            print 'Calling update_precipitable_water_content()...'

        #------------------------------------------------------------
        # Notes:  W_p is precipitable water content in centimeters,
        #         which depends on air temp and relative humidity.
        #------------------------------------------------------------
        arg      = numpy.float64( 0.0614 * self.T_dew )
        self.W_p = numpy.float64(1.12) * numpy.exp( arg )  # [cm]

    #   update_precipitable_water_content()
    #-------------------------------------------------------------------
    def update_latent_heat_flux(self):

        if (self.DEBUG):
            print 'Calling update_latent_heat_flux()...'

        #--------------------------------------------------------
        # Notes:  Pressure units cancel out because e_air and
        #         e_surf (in numer) have same units (mbar) as
        #         p0 (in denom).
        #--------------------------------------------------------        
        # According to Dingman (2002, p. 273), constant should
        # be 0.622 instead of 0.662 (Zhang et al., 2000).
        #--------------------------------------------------------
        # const = float64(0.622)
        const   = float64(0.662)
        factor  = (self.rho_air * self.Lv * self.De)
        delta_e = (self.e_air - self.e_surf)
        self.Qe = factor * delta_e * (const / self.p0)

    #   update_latent_heat_flux()
    #-------------------------------------------------------------------
    def update_conduction_heat_flux(self):

        if (self.DEBUG):
            print 'Calling update_conduction_heat_flux()...'

        #-----------------------------------------------------------------
        # Notes: The conduction heat flux from snow to soil for computing
        #        snowmelt energy, Qm, is close to zero.

        #        However, the conduction heat flux from surface and sub-
        #        surface for computing Qet is given by Fourier's Law,
        #        namely Qc = Ks(Tx - Ts)/x.

        #        All the Q's have units of W/m^2 = J/(m^2 s).
        #-----------------------------------------------------------------
        self.Qc = numpy.float64(0)

    #   update_conduction_heat_flux()
    #-------------------------------------------------------------------
    def update_advection_heat_flux(self):

        if (self.DEBUG):
            print 'Calling update_advection_heat_flux()...'

        #------------------------------------------------------
        # Notes: All the Q's have units of W/m^2 = J/(m^2 s).
        #------------------------------------------------------
        self.Qa = numpy.float64(0)
        
    #   update_advection_heat_flux()
    #-------------------------------------------------------------------
    def update_julian_day(self):

        if (self.DEBUG):
            print 'Calling update_julian_day()...'

        #----------------------------------
        # Update the *decimal* Julian day
        #----------------------------------
        self.julian_day += (self.dt / self.secs_per_day) # [days]
  
        #------------------------------------------
        # Compute the offset from True Solar Noon
        # clock_hour is in 24-hour military time
        # but it can have a decimal part.
        #------------------------------------------
        dec_part   = self.julian_day - numpy.int16(self.julian_day)
        clock_hour = dec_part * self.hours_per_day
        ## print '    Computing solar_noon...'
        solar_noon = solar.True_Solar_Noon( self.julian_day,
                                            self.lon_deg,
                                            self.GMT_offset )
        ## print '    Computing TSN_offset...'
        self.TSN_offset = (clock_hour - solar_noon)    # [hours]
    
    #   update_julian_day()
    #-------------------------------------------------------------------
    def update_net_shortwave_radiation(self):

        #---------------------------------------------------------
        # Notes:  If time is before local sunrise or after local
        #         sunset then Qn_SW should be zero.
        #---------------------------------------------------------
        if (self.DEBUG):
            print 'Calling update_net_shortwave_radiation()...'

        #--------------------------------
        # Compute Qn_SW for this time
        #--------------------------------
        Qn_SW = solar.Clear_Sky_Radiation( self.lat_deg,
                                           self.julian_day,
                                           self.W_p,
                                           self.TSN_offset,
                                           self.alpha,
                                           self.beta,
                                           self.albedo,
                                           self.dust_atten )

        self.Qn_SW = Qn_SW
        ## self.Qn_SW = Qn_SW * self.Qsw_prefactor

        #--------------
        # For testing
        #--------------
        # self.Qn_SW = numpy.float64(100)    # [W/m^2]
        
    #   update_net_shortwave_radiation()
    #-------------------------------------------------------------------
    def update_em_air(self):

        if (self.DEBUG):
            print 'Calling update_em_air()...'

        #---------------------------------------------------------
        # NB!  The Brutsaert and Satterlund formulas for air
        #      emissivity as a function of air temperature are in
        #      close agreement; see compare_em_air_methods().
        #      However, we must pay close attention to whether
        #      equations require units of kPa, Pa, or mbar.
        #
        #             100 kPa = 1 bar = 1000 mbars
        #                => 1 kPa = 10 mbars
        #---------------------------------------------------------
        # NB!  Temperatures are assumed to be given with units
        #      of degrees Celsius and are converted to Kelvin
        #      wherever necessary by adding C_to_K = 273.15.
        #
        #      RH = relative humidity [unitless]
        #---------------------------------------------------------
        # NB!  I'm not sure about how F is added at end because
        #      of how the equation is printed in Dingman (2002).
        #      But it reduces to other formulas as it should.
        #---------------------------------------------------------
        T_air_K = self.T_air + self.C_to_K
        
        if not(self.SATTERLUND):
            #-----------------------------------------------------
            # Brutsaert (1975) method for computing emissivity
            # of the air, em-air.  This formula uses e_air with
            # units of kPa. (From Dingman (2002, p. 196).)
            # See notes for update_vapor_pressure().
            #-----------------------------------------------------
            e_air_kPa = self.e_air / numpy.float64(10)  # [kPa]
            F       = self.canopy_factor
            C       = self.cloud_factor
            term1 = (1.0 - F) * 1.72 * (e_air_kPa / T_air_K) ** self.one_seventh
            term2 = (1.0 + (0.22 * C ** 2.0))
            self.em_air = (term1 * term2) + F
        else:
            #--------------------------------------------------------
            # Satterlund (1979) method for computing the emissivity
            # of the air, em_air, that is intended to "correct
            # apparent deficiencies in this formulation at air
            # temperatures below 0 degrees C" (see G. Liston)
            # Liston cites Aase and Idso(1978), Satterlund (1979)
            #--------------------------------------------------------
            e_air_mbar  = self.e_air
            eterm  = numpy.exp(-1 * (e_air_mbar)**(T_air_K / 2016) )
            self.em_air = 1.08 * (1.0 - eterm)
               
    #   update_em_air()    
    #-------------------------------------------------------------------
    def update_net_longwave_radiation(self):

        #----------------------------------------------------------------
        # Notes: Net longwave radiation is computed using the
        #        Stefan-Boltzman law.  All four data types
        #        should be allowed (scalar, time series, grid or
        #        grid stack).
        #
        #        Qn_LW = (LW_in - LW_out)
        #        LW_in   = em_air  * sigma * (T_air  + 273.15)^4
        #        LW_out  = em_surf * sigma * (T_surf + 273.15)^4
        #
        #        Temperatures in [deg_C] must be converted to
        #        [deg_K].  Recall that absolute zero occurs at
        #        0 [deg_K] or -273.15 [deg_C].
        #
        #----------------------------------------------------------------
        # First, e_air is computed as:
        #   e_air = RH * 0.611 * exp[(17.3 * T_air) / (T_air + 237.3)]
        # Then, em_air is computed as:
        #   em_air = (1 - F) * 1.72 * [e_air / (T_air + 273.15)]^(1/7) *
        #             (1 + 0.22 * C^2) + F
        #----------------------------------------------------------------
        if (self.DEBUG):
            print 'Calling update_net_longwave_radiation()...'

        #--------------------------------
        # Compute Qn_LW for this time
        #--------------------------------
        T_air_K  = self.T_air  + self.C_to_K
        T_surf_K = self.T_surf + self.C_to_K
        LW_in    = self.em_air  * self.sigma * (T_air_K)** 4.0
        LW_out   = self.em_surf * self.sigma * (T_surf_K)** 4.0
        LW_out   = LW_out + ((1.0 - self.em_surf) * LW_in)
               
        self.Qn_LW = (LW_in - LW_out)

        #--------------
        # For testing
        #--------------
        # self.Qn_LW = numpy.float64(10)    # [W/m^2]
        
    #   update_net_longwave_radiation()
    #-------------------------------------------------------------------
    def update_net_energy_flux(self):

        if (self.DEBUG):
            print 'Calling update_net_energy_flux()...'

        #------------------------------------------------------
        # Notes: Q_sum is used by "snow_energy_balance.py".
        #------------------------------------------------------
        #        Qm    = energy used to melt snowpack (if > 0)
        #        Qn_SW = net shortwave radiation flux (solar)
        #        Qn_LW = net longwave radiation flux (air, surface)
        #        Qh    = sensible heat flux from turbulent convection
        #                between snow surface and air
        #        Qe    = latent heat flux from evaporation, sublimation,
        #                and condensation
        #        Qa    = energy advected by moving water (i.e. rainfall)
        #                (ARHYTHM assumes this to be negligible; Qa=0.)
        #        Qc    = energy flux via conduction from snow to soil
        #                (ARHYTHM assumes this to be negligible; Qc=0.)
        #        Ecc   = cold content of snowpack = amount of energy
        #                needed before snow can begin to melt [J/m^2]

        #        All Q's here have units of [W/m^2].
        #        Are they all treated as positive quantities ?

        #        rho_air  = density of air [kg/m^3]
        #        rho_snow = density of snow [kg/m^3]
        #        Cp_air   = specific heat of air [J/kg/deg_C]
        #        Cp_snow  = heat capacity of snow [J/kg/deg_C]
        #                 = ???????? = specific heat of snow
        #        Kh       = eddy diffusivity for heat [m^2/s]
        #        Ke       = eddy diffusivity for water vapor [m^2/s]
        #        Lv       = latent heat of vaporization [J/kg]
        #        Lf       = latent heat of fusion [J/kg]
        #        ------------------------------------------------------
        #        Dn       = bulk exchange coeff for the conditions of
        #                   neutral atmospheric stability [m/s]
        #        Dh       = bulk exchange coeff for heat
        #        De       = bulk exchange coeff for vapor
        #        ------------------------------------------------------
        #        T_air    = air temperature [deg_C]
        #        T_surf   = surface temperature [deg_C]
        #        T_snow   = average snow temperature [deg_C]
        #        RH       = relative humidity [unitless] (in [0,1])
        #        e_air    = air vapor pressure at height z [mbar]
        #        e_surf   = surface vapor pressure [mbar]
        #        ------------------------------------------------------
        #        h_snow   = snow depth [m]
        #        z        = height where wind speed is uz [m]
        #        uz       = wind speed at height z [m/s]
        #        p0       = atmospheric pressure [mbar]
        #        T0       = snow temperature when isothermal [deg_C]
        #                   (This is usually 0.)
        #        z0_air   = surface roughness length scale [m]
        #                   (includes vegetation not covered by snow)
        #                   (Values from page 1033: 0.0013, 0.02 [m])
        #        kappa    = von Karman's constant [unitless] = 0.41
        #        dt       = snowmelt timestep [seconds]
        #----------------------------------------------------------------
        self.Q_sum = self.Qn_SW + self.Qn_LW + self.Qh + \
                     self.Qe + self.Qa + self.Qc    #[W/m^2]

    #   update_net_energy_flux()   
    #-------------------------------------------------------------------  
    def open_input_files(self):

        if (self.DEBUG):
            print 'Calling open_input_files()...'
       
        self.P_file      = self.in_directory + self.P_file
        self.T_air_file  = self.in_directory + self.T_air_file
        self.T_surf_file = self.in_directory + self.T_surf_file
        self.RH_file     = self.in_directory + self.RH_file
        self.p0_file     = self.in_directory + self.p0_file
        self.uz_file     = self.in_directory + self.uz_file
        self.z_file      = self.in_directory + self.z_file
        self.z0_air_file = self.in_directory + self.z0_air_file

        self.albedo_file        = self.in_directory + self.albedo_file
        self.em_surf_file       = self.in_directory + self.em_surf_file
        self.dust_atten_file    = self.in_directory + self.dust_atten_file
        self.cloud_factor_file  = self.in_directory + self.cloud_factor_file
        self.canopy_factor_file = self.in_directory + self.canopy_factor_file

        self.P_unit      = model_input.open_file(self.P_type,      self.P_file)
        self.T_air_unit  = model_input.open_file(self.T_air_type,  self.T_air_file)
        self.T_surf_unit = model_input.open_file(self.T_surf_type, self.T_surf_file)
        self.RH_unit     = model_input.open_file(self.RH_type,     self.RH_file)
        self.p0_unit     = model_input.open_file(self.p0_type,     self.p0_file)
        self.uz_unit     = model_input.open_file(self.uz_type,     self.uz_file)
        self.z_unit      = model_input.open_file(self.z_type,      self.z_file)
        self.z0_air_unit = model_input.open_file(self.z0_air_type, self.z0_air_file)
               
        #-----------------------------------------------
        # These are needed to compute Qn_SW and Qn_LW.
        #-----------------------------------------------
        self.albedo_unit        = model_input.open_file(self.albedo_type,
                                                        self.albedo_file)
        self.em_surf_unit       = model_input.open_file(self.em_surf_type,
                                                        self.em_surf_file)
        self.dust_atten_unit    = model_input.open_file(self.dust_atten_type,
                                                        self.dust_atten_file)
        self.cloud_factor_unit  = model_input.open_file(self.cloud_factor_type,
                                                        self.cloud_factor_file)
        self.canopy_factor_unit = model_input.open_file(self.canopy_factor_type,
                                                        self.canopy_factor_file)
        #----------------------------------------------------------------------------
        # Note: GMT_offset plus slope and aspect grids will be read separately.
        #----------------------------------------------------------------------------

##        self.Qn_SW_unit  = model_input.open_file(self.Qn_SW_type,  self.Qn_SW_file)
##        self.Qn_LW_unit  = model_input.open_file(self.Qn_LW_type,  self.Qn_LW_file)
        
    #   open_input_files()
    #-------------------------------------------------------------------  
    def read_input_files(self):

        if (self.DEBUG):
            print 'Calling read_input_files()...'

        rti = self.rti

        #--------------------------------------------------------
        # All grids are assumed to have a data type of Float32.
        #--------------------------------------------------------
        # NB! read_next() returns None if TYPE arg is "Scalar".
        #--------------------------------------------------------
        P = model_input.read_next(self.P_unit, self.P_type, rti,
                                  factor=self.mmph_to_mps)

        if (P != None):
            self.P = numpy.float64(P)  ##### (or float32 ??) #####

            if (self.DEBUG or (self.time_index == 0)):
                print 'In read_input_files():'
                print '   min(P) =', P.min() * self.mps_to_mmph, ' [mmph]'
                print '   max(P) =', P.max() * self.mps_to_mmph, ' [mmph]'
                print ' '
        else:
            #-----------------------------------------------
            # Either self.P_type is "Scalar" or we've read
            # all of the data in the rain_rates file.
            #-----------------------------------------------
            if (self.P_type.lower() != 'scalar'):
                #------------------------------------
                # Precip is unique in this respect.
                #------------------------------------
                self.P = numpy.float64(0)
                if (self.DEBUG):
                    print 'Reached end of file:', self.P_file
                    print '  P set to 0 by read_input_files().'
            elif (self.time_sec >= self.dt):
                self.P = numpy.float64(0)
                if (self.DEBUG):
                    print 'Reached end of scalar rainfall DT.'
                    print '  P set to 0.'
                    ## print 'time_sec =', self.time_sec
                    ## print 'met dt   =', self.dt
                    
##        print '######### In met_base.read_input_files() #######'
##        print 'self.P_type =', self.P_type
##        print 'self.P      =', self.P
        
        T_air = model_input.read_next(self.T_air_unit, self.T_air_type, rti)
        if (T_air != None): self.T_air = T_air

        T_surf = model_input.read_next(self.T_surf_unit, self.T_surf_type, rti)
        if (T_surf != None): self.T_surf = T_surf

        RH = model_input.read_next(self.RH_unit, self.RH_type, rti)
        if (RH != None): self.RH = RH

        p0 = model_input.read_next(self.p0_unit, self.p0_type, rti)
        if (p0 != None): self.p0 = p0

        uz = model_input.read_next(self.uz_unit, self.uz_type, rti)
        if (uz != None): self.uz = uz

        z = model_input.read_next(self.z_unit, self.z_type, rti)
        if (z != None): self.z = z

        z0_air = model_input.read_next(self.z0_air_unit, self.z0_air_type, rti)
        if (z0_air != None): self.z0_air = z0_air

        #----------------------------------------------------------------------------
        # These are needed to compute Qn_SW and Qn_LW.
        #----------------------------------------------------------------------------
        # Note: We could later write a version of read_next() that takes "self"
        #       and "var_name" as args and that uses "exec()".
        #----------------------------------------------------------------------------
        albedo = model_input.read_next(self.albedo_unit, self.albedo_type, rti)
        if (albedo != None): self.albedo = albedo

        em_surf = model_input.read_next(self.em_surf_unit, self.em_surf_type, rti)
        if (em_surf != None): self.em_surf = em_surf

        dust_atten = model_input.read_next(self.dust_atten_unit, self.dust_atten_type, rti)
        if (dust_atten != None): self.dust_atten = dust_atten

        cloud_factor = model_input.read_next(self.cloud_factor_unit, self.cloud_factor_type, rti)
        if (cloud_factor != None): self.cloud_factor = cloud_factor

        canopy_factor = model_input.read_next(self.canopy_factor_unit, self.canopy_factor_type, rti)
        if (canopy_factor != None): self.canopy_factor = canopy_factor

        #-------------------------------------------------------------
        # Compute Qsw_prefactor from cloud_factor and canopy factor.
        #-------------------------------------------------------------
        ## self.Qsw_prefactor = 
        
        #-------------------------------------------------------------
        # These are currently treated as input data, but are usually
        # generated by functions in Qnet_file.py.  Later on, we'll
        # provide the option to compute them "on the fly" with new
        # functions called "update_net_shortwave_radiation()" and
        # "update_net_longwave_radiation()", called from update().
        #-------------------------------------------------------------        
##        Qn_SW = model_input.read_next(self.Qn_SW_unit, self.Qn_SW_type, rti)
##        if (Qn_SW != None): self.Qn_SW = Qn_SW
##
##        Qn_LW = model_input.read_next(self.Qn_LW_unit, self.Qn_LW_type, rti)
##        if (Qn_LW != None): self.Qn_LW = Qn_LW
         
    #   read_input_files()
    #-------------------------------------------------------------------  
    def close_input_files(self):

        if (self.DEBUG):
            print 'Calling close_input_files()...'

        if (self.P_type      != 'Scalar'): self.P_unit.close()
        if (self.T_air_type  != 'Scalar'): self.T_air_unit.close()
        if (self.T_surf_type != 'Scalar'): self.T_surf_unit.close()
        if (self.RH_type     != 'Scalar'): self.RH_unit.close()
        if (self.p0_type     != 'Scalar'): self.p0_unit.close()
        if (self.uz_type     != 'Scalar'): self.uz_unit.close()
        if (self.z_type      != 'Scalar'): self.z_unit.close()
        if (self.z0_air_type != 'Scalar'): self.z0_air_unit.close()

        #---------------------------------------------------
        # These are needed to compute Qn_SW and Qn_LW.
        #---------------------------------------------------
        if (self.albedo_type        != 'Scalar'): self.albedo_unit.close()
        if (self.em_surf_type       != 'Scalar'): self.em_surf_unit.close()
        if (self.dust_atten_type    != 'Scalar'): self.dust_atten_unit.close()
        if (self.cloud_factor_type  != 'Scalar'): self.cloud_factor_unit.close()
        if (self.canopy_factor_type != 'Scalar'): self.canopy_factor_unit.close()
        
##        if (self.Qn_SW_type  != 'Scalar'): self.Qn_SW_unit.close()        
##        if (self.Qn_LW_type  != 'Scalar'): self.Qn_LW_unit.close()
        
##        if (self.P_file      != ''): self.P_unit.close()
##        if (self.T_air_file  != ''): self.T_air_unit.close()
##        if (self.T_surf_file != ''): self.T_surf_unit.close()
##        if (self.RH_file     != ''): self.RH_unit.close()
##        if (self.p0_file     != ''): self.p0_unit.close()
##        if (self.uz_file     != ''): self.uz_unit.close()
##        if (self.z_file      != ''): self.z_unit.close()
##        if (self.z0_air_file != ''): self.z0_air_unit.close()
##        #--------------------------------------------------------
##        if (self.Qn_SW_file  != ''): self.Qn_SW_unit.close()        
##        if (self.Qn_LW_file  != ''): self.Qn_LW_unit.close()
        
    #   close_input_files()
    #-------------------------------------------------------------------  
    def update_outfile_names(self):

        if (self.DEBUG):
            print 'Calling update_outfile_names()...'
        
        #-------------------------------------------------
        # Notes:  Append out_directory to outfile names.
        #-------------------------------------------------
        self.ea_gs_file  = (self.out_directory + self.ea_gs_file  )
        self.es_gs_file  = (self.out_directory + self.es_gs_file  )
        self.Qsw_gs_file = (self.out_directory + self.Qsw_gs_file )
        self.Qlw_gs_file = (self.out_directory + self.Qlw_gs_file )
        self.ema_gs_file = (self.out_directory + self.ema_gs_file )
        #------------------------------------------------------------
        self.ea_ts_file  = (self.out_directory + self.ea_ts_file  )
        self.es_ts_file  = (self.out_directory + self.es_ts_file  )
        self.Qsw_ts_file = (self.out_directory + self.Qsw_ts_file )
        self.Qlw_ts_file = (self.out_directory + self.Qlw_ts_file )
        self.ema_ts_file = (self.out_directory + self.ema_ts_file )
        
##        self.ea_gs_file = (self.case_prefix + '_2D-ea.rts')
##        self.es_gs_file = (self.case_prefix + '_2D-es.rts')
##        #-----------------------------------------------------
##        self.ea_ts_file = (self.case_prefix + '_0D-ea.txt')
##        self.es_ts_file = (self.case_prefix + '_0D-es.txt')

    #   update_outfile_names()   
    #-------------------------------------------------------------------  
    def open_output_files(self):

        if (self.DEBUG):
            print 'Calling open_output_files()...'
        NIO_OK = model_output.check_nio_package()
        self.update_outfile_names()
        
        #--------------------------------------
        # Open new files to write grid stacks
        #--------------------------------------
        if (self.SAVE_EA_GRIDS):
            model_output.open_new_gs_file( self, self.ea_gs_file, self.rti,
                                           ## var_name='e_air',
                                           var_name='ea',
                                           long_name='vapor_pressure_in_air',
                                           units_name='mbar')
            
        if (self.SAVE_ES_GRIDS):
            model_output.open_new_gs_file( self, self.es_gs_file, self.rti,
                                           ## var_name='e_surf',
                                           var_name='es',
                                           long_name='vapor_pressure_at_surface',
                                           units_name='mbar')
        if (self.SAVE_QSW_GRIDS):
            model_output.open_new_gs_file( self, self.Qsw_gs_file, self.rti,
                                           var_name='Qsw',
                                           long_name='net_shortwave_radiation',
                                           units_name='W/m^2')
            
        if (self.SAVE_QLW_GRIDS):
            model_output.open_new_gs_file( self, self.Qlw_gs_file, self.rti,
                                           var_name='Qlw',
                                           long_name='net_longwave_radiation',
                                           units_name='W/m^2')
        if (self.SAVE_EMA_GRIDS):
            model_output.open_new_gs_file( self, self.ema_gs_file, self.rti,
                                           var_name='ema',
                                           long_name='air_emissivity',
                                           units_name='none')
            
        #--------------------------------------
        # Open new files to write time series
        #--------------------------------------
        IDs = self.outlet_IDs 
        if (self.SAVE_EA_PIXELS):
            model_output.open_new_ts_file( self, self.ea_ts_file, IDs,
                                           ## var_name='e_air',
                                           var_name='ea',
                                           long_name='vapor_pressure_in_air',
                                           units_name='mbar')

        if (self.SAVE_ES_PIXELS):
            model_output.open_new_ts_file( self, self.es_ts_file, IDs,
                                           ## var_name='e_surf',
                                           var_name='es',
                                           long_name='vapor_pressure_at_surface',
                                           units_name='mbar')

        if (self.SAVE_QSW_PIXELS):
            model_output.open_new_ts_file( self, self.Qsw_ts_file, IDs,
                                           var_name='Qsw',
                                           long_name='net_shortwave_radiation',
                                           units_name='W/m^2')

        if (self.SAVE_QLW_PIXELS):
            model_output.open_new_ts_file( self, self.Qlw_ts_file, IDs,
                                           var_name='Qlw',
                                           long_name='net_longwave_radiation',
                                           units_name='W/m^2')
            
        if (self.SAVE_EMA_PIXELS):
            model_output.open_new_ts_file( self, self.ema_ts_file, IDs,
                                           var_name='ema',
                                           long_name='air_emissivity',
                                           units_name='none')
            
    #   open_output_files()
    #-------------------------------------------------------------------
    def write_output_files(self, time_seconds=None):

        if (self.DEBUG):
            print 'Calling write_output_files()...'

        #-----------------------------------------
        # Allows time to be passed from a caller
        #-----------------------------------------
        if (time_seconds is None):
            time_seconds = self.time_sec
        model_time = int(time_seconds)
        
        #----------------------------------------
        # Save computed values at sampled times
        #----------------------------------------
        if (model_time % int(self.save_grid_dt) == 0):
            self.save_grids()
        if (model_time % int(self.save_pixels_dt) == 0):
            self.save_pixel_values()

    #  write_output_files()
    #-------------------------------------------------------------------
    def close_output_files(self):
    
        if (self.SAVE_EA_GRIDS):   model_output.close_gs_file( self, 'ea')
        if (self.SAVE_ES_GRIDS):   model_output.close_gs_file( self, 'es')
        if (self.SAVE_QSW_GRIDS):  model_output.close_gs_file( self, 'Qsw')
        if (self.SAVE_QLW_GRIDS):  model_output.close_gs_file( self, 'Qlw')
        if (self.SAVE_EMA_GRIDS):  model_output.close_gs_file( self, 'ema')
        #-------------------------------------------------------------------
        if (self.SAVE_EA_PIXELS):  model_output.close_ts_file( self, 'ea') 
        if (self.SAVE_ES_PIXELS):  model_output.close_ts_file( self, 'es') 
        if (self.SAVE_QSW_PIXELS): model_output.close_ts_file( self, 'Qsw') 
        if (self.SAVE_QLW_PIXELS): model_output.close_ts_file( self, 'Qlw')
        if (self.SAVE_EMA_PIXELS): model_output.close_ts_file( self, 'ema')
        
    #   close_output_files()        
    #-------------------------------------------------------------------  
    def save_grids(self):
       
        if (self.SAVE_EA_GRIDS):
            model_output.add_grid( self, self.e_air,  'ea', self.time_min )
            
        if (self.SAVE_ES_GRIDS):
            model_output.add_grid( self, self.e_surf, 'es', self.time_min )

        if (self.SAVE_QSW_GRIDS):
            model_output.add_grid( self, self.Qn_SW, 'Qsw', self.time_min )
            
        if (self.SAVE_QLW_GRIDS):
            model_output.add_grid( self, self.Qn_LW, 'Qlw', self.time_min )

        if (self.SAVE_EMA_GRIDS):
            model_output.add_grid( self, self.em_air, 'ema', self.time_min )
            
    #   save_grids()            
    #-------------------------------------------------------------------  
    def save_pixel_values(self):

        IDs  = self.outlet_IDs
        time = self.time_min   ######
        
        if (self.SAVE_EA_PIXELS):
            model_output.add_values_at_IDs( self, time, self.e_air,  'ea', IDs )
            
        if (self.SAVE_ES_PIXELS):
            model_output.add_values_at_IDs( self, time, self.e_surf, 'es', IDs )

        if (self.SAVE_QSW_PIXELS):
            model_output.add_values_at_IDs( self, time, self.Qn_SW, 'Qsw', IDs )
            
        if (self.SAVE_QLW_PIXELS):
            model_output.add_values_at_IDs( self, time, self.Qn_LW, 'Qlw', IDs )

        if (self.SAVE_EMA_PIXELS):
            model_output.add_values_at_IDs( self, time, self.em_air, 'ema', IDs )
            
    #   save_pixel_values()
    #-------------------------------------------------------------------
#---------------------------------------------------------------------------------
def compare_em_air_methods():

    #--------------------------------------------------------------
    # Notes:  There are two different methods that are commonly
    #         used to compute the vapor pressure of air, e_air,
    #         and then the emissivity of air, em_air, for use in
    #         longwave radiation calculations.  This routine
    #         compares them graphically.
    #
    # NB!     This hasn't been tested since conversion from IDL.
    #-------------------------------------------------------------
    import matplotlib.pyplot
    
    T_air = arange(80, dtype='Float32') - float32(40.0)   #[Celsius]  (-40 to 40)
    RH  = float32(1.0)
    C2K = float32(273.15)
    
    #--------------------------
    # Brutsaert (1975) method
    #--------------------------
    e_air1  = RH * float32(0.611) * exp((float32(17.3) * T_air) / (T_air + float32(237.3)))  # [kPa]
    em_air1 = float32(1.72) * (e_air1 / (T_air + C2K)) ** (float32(1.0) / 7)
    
    #---------------------------
    # Satterlund (1979) method
    #----------------------------
    # NB! e_air has units of Pa
    #----------------------------
    e_air2  = RH * float64(10) ** (float32(11.40) - (float64(2353) / (T_air + C2K)))   # [Pa]
    eterm   = exp(-float64(1) * (e_air2 / float64(100)) ** ((T_air + C2K) / float64(2016)))
    em_air2 = float64(1.08) * (float64(1) - eterm)
    
    #----------------------------
    # Plot the two e_air curves
    #--------------------------------
    # These two agree quite closely
    #--------------------------------
    matplotlib.pyplot.figure(figsize=(8, 6), dpi=80)
    matplotlib.pyplot.show()
    matplotlib.pyplot.plot(T_air, e_air1)
    matplotlib.pyplot.show()
    ## oplot(T_air, (e_air2 / float32(1000.0)), psym=-3)   # [Pa -> kPa]
    
    #-----------------------------
    # Plot the two em_air curves
    #--------------------------------------------------
    # These two don't agree very well for some reason
    #--------------------------------------------------
    matplotlib.pyplot.figure(figsize=(8, 6), dpi=80)
    matplotlib.pyplot.show()
    matplotlib.pyplot.plot(T_air, em_air1)
    matplotlib.pyplot.show()
    ## oplot(T_air, em_air2, psym=-3)
    
#   compare_em_air_Methods
#---------------------------------------------------------------------------------
    
        
