
## Copyright (c) 2001-2010, Scott D. Peckham
## January 2009  (converted from IDL)
## July, August 2009
## May 2010 (changes to unit_test() and read_cfg_file()

#-----------------------------------------------------------------------
#  NOTES:  This file defines an Energy-Balance snowmelt component
#          and related functions.  It inherits from the snowmelt
#          "base class" in "snow_base.py".
#-----------------------------------------------------------------------
#
#  unit_test()
#
#  class snow_energy_balance
#
#      get_attribute()        # (10/26/11)
#      check_input_types()
#      update_meltrate()
#
#-----------------------------------------------------------------------

from numpy import *
import numpy

import snow_base
import tf_utils

#-----------------------------------------------------------------------
def unit_test():

    c = snow_component()
    c.CCA   = False
    c.DEBUG = False
    
    #-----------------------------------------
    # This function adjusts for the platform
    # and can be changed in "tf_utils.py".
    #-----------------------------------------
    cfg_directory = tf_utils.TF_Test_Directory()
    cfg_prefix    = tf_utils.TF_Test_Case_Prefix()
    c.site_prefix = 'Treynor'  #############    

    #-------------------------------------------
    # Call initialize() and call update() once
    #-------------------------------------------
##    print 'STATUS =', c.get_status()
##    c.initialize( mode="driver" )
##    print 'STATUS =', c.get_status()
##    time_sec = float64(0)
##    c.update(time_sec)
##    print 'STATUS =', c.get_status()

    c.run_model( cfg_directory=cfg_directory,
                 cfg_prefix=cfg_prefix )
    
#   unit_test()
#-----------------------------------------------------------------------
class snow_component(snow_base.snow_component):

    #-------------------------------------------------------------------
    def get_attribute(self, att_name):

        map = {'comp_name':          'SnowEnergyBalance',
               'version':            '3.1',
               'model_name':         'Snowmelt_Energy_Balance',    # (Snowmelt ?)
               'model_family':       'TopoFlow',
               'cfg_template_file':  'Snow_Energy_Balance.cfg.in',
               'cfg_extension':      '_snow_energy_balance.cfg',
               'cmt_var_prefix':     '/SnowEnergyBalance/Input/Var/',
               'gui_xml_file':       '/home/csdms/cca/topoflow/3.1/src/share/cmt/gui/Snow_Energy_Balance.xml',
               'dialog_title':       'Snowmelt: Energy Balance Parameters',  # (Snowmelt ?)
               'time_step_type':     'fixed',
               'time_units':         'seconds',
               'mesh_type':          'none',
               'author_name':        'Scott Peckham'}

        try:
            return map[ att_name.lower() ]
        except:
            print '###################################################'
            print ' ERROR: Could not find attribute: ' + att_name
            print '###################################################'
            print ' '

    #   get_attribute() 
    #-------------------------------------------------------------------
    def check_input_types(self):

        #--------------------------------------------------
        # Notes: rho_H2O, Cp_snow, rho_air and Cp_air are
        #        currently always scalars.
        #--------------------------------------------------        
        are_scalars = numpy.array([
##                          self.pp.is_scalar('rate'),
##                          self.pp.is_scalar('duration'),
                          #--------------------------------
                          self.mp.is_scalar('P'),
                          self.mp.is_scalar('rho_H2O'),
                          self.mp.is_scalar('rho_air'),
                          self.mp.is_scalar('Cp_air'),
                          self.mp.is_scalar('T_air'),
                          self.mp.is_scalar('T_surf'),
                          self.mp.is_scalar('RH'),
                          self.mp.is_scalar('p0'),
                          self.mp.is_scalar('uz'),
                          self.mp.is_scalar('z'),
                          self.mp.is_scalar('z0_air'),
                          #--------------------------------
                          self.mp.is_scalar('Qn_SW'),
                          self.mp.is_scalar('Qn_LW'),
                          #--------------------------------
                          self.is_scalar('rho_snow'),
                          self.is_scalar('Cp_snow'),
                          self.is_scalar('h0_snow'),
                          self.is_scalar('h0_swe') ])


        self.ALL_SCALARS = numpy.all(are_scalars)
        
    #   check_input_types()
    #-------------------------------------------------------------------
    def update_meltrate(self):

        #------------------------------------------------------------
        # Notes: See notes in "met_base.py" for the method called
        #        "update_net_energy_flux()".

        #        This version uses "Q_sum" which is computed as a
        #        state variable for a meteorology component
        #        (e.g. met_base.py).

        #        Arguments are assumed to be references to a scalar
        #            or grid so that:

        #        M  = water equivalent of snowmelt [m/s]
        #        M_max  = max possible meltrate if all snow melts
        #        T_air  = air temperature [deg_C]

        #        Model must start when snow is isothermal. (CHECK)
        #        Cooling of the snowpack is not considered.

        #        86400000d = 1000 [mm/m] * 60 [sec/min] *
        #                    60 [min/sec] * 24 [hrs/day]

        #        rho_snow is not needed in formula here, but is
        #        needed to convert snowmelt to water equivalent?
        #-------------------------------------------------------------
     
        #----------------------------------
        # Compute energy-balance meltrate   
        #------------------------------------------------------
        # Ecc is initialized with the Initial_Cold_Content
        # function by Initialize_Snow_Vars function (2/21/07)
        #------------------------------------------------------
        # The following pseudocode only works for scalars but
        # is otherwise equivalent to that given below and
        # clarifies the logic:
        #------------------------------------------------------
        #  if (Q_sum gt 0) then begin
        #      if ((Q_sum * dt) gt Ecc) then begin
        #          ;-----------------------------------------
        #          ;Snow is melting.  Use some of Q_sum to
        #          ;overcome Ecc, and remainder to melt snow
        #          ;-----------------------------------------
        #          Qm  = Q_sum - (Ecc/dt)
        #          Ecc = 0
        #          M   = (Qm / (rho_w * Lf))
        #      endif else begin
        #          ;----------------------------
        #          ;Snow is warming; reduce Ecc
        #          ;----------------------------
        #          Ecc = (Ecc - (Q_sum * dt))
        #          M   = 0d
        #      endelse
        #  endif else begin
        #      ;------------------------------
        #      ;Snow is cooling; increase Ecc
        #      ;------------------------------
        #      Ecc = Ecc - (Q_sum * dt)
        #      M   = 0d
        #  endelse
        #-------------------------------------------------------
        # Q_sum = Qn_SW + Qn_LW + Qh + Qe + Qa + Qc      #[W/m^2]
        Q_sum = self.get_port_data('Q_sum', self.mp)    #[W/m^2]
        Qcc   = (self.Ecc / self.dt)                    #[W/m^2]
        Qm    = maximum((Q_sum - Qcc), float64(0))                   #[W/m^2]
        Ecc   = maximum((self.Ecc - (Q_sum * self.dt)), float64(0))  #[J/m^2]
        
        #print 'Qm = ', Qm
        #print ' '
        
        #-------------------------------------
        # Convert melt energy to a melt rate
        #------------------------------------------
        # Lf = latent heat of fusion [J/kg]
        # Lv = latent heat of vaporization [J/kg]
        # M  = (Qm/ (rho_w * Lf))
        #------------------------------------------
        # rho_w = 1000d       ;[kg/m^3]
        # Lf    = 334000d     ;[J/kg = W*s/kg]
        # So (rho_w * Lf) = 3.34e+8  [J/m^3]
        #------------------------------------------
        M       = (Qm / float64(3.34E+8))   #[m/s]
        self.SM = maximum(M, float64(0))
   
        #-------------------------------------------------------
        # Note: enforce_max_meltrate() method is always called
        #       by the base class to make sure that meltrate
        #       does not exceed the max possible.
        #-------------------------------------------------------
        #  self.enforce_max_meltrate()
            
    #   update_meltrate()
    #-------------------------------------------------------------------
    
#-------------------------------------------------------------------------  
###-------------------------------------------------------------------------
##def Energy_Balance_Meltrate(Qn_SW, Qn_LW, T_air, T_surf, RH, p0, \
##                            uz, z, z0_air, rho_air, Cp_air, Ecc, \
##                            h_snow, rho_snow, Cp_snow, dt, \
##                            e_air, e_surf):  #(returned)
##
##    #-----------------------------------------------------------------
##    #Notes:  3/13/07.  This function used to have vapor pressure
##    #        arguments e_air and e_surf.  However, e_air is always
##    #        computed as a function of T_air and RH and e_surf is
##    #        computed as a function of T_surf (saturated vap. press.)
##    #        So it makes sense to remove these two arguments and add
##    #        RH (relative humidity).  This change only affects the
##    #        Latent_Heat_Flux function call, which calls a new
##    #        function called Vapor_Pressure.
##    #-----------------------------------------------------------------
##    #        Qm    = energy used to melt snowpack (if > 0)
##    #        Qn_SW = net shortwave radiation flux (solar)
##    #        Qn_LW = net longwave radiation flux (air, surface)
##    #        Qh    = sensible heat flux from turbulent convection
##    #                between snow surface and air
##    #        Qe    = latent heat flux from evaporation, sublimation,
##    #                and condensation
##    #        Qa    = energy advected by moving water (i.e. rainfall)
##    #                (ARHYTHM assumes this to be negligible; Qa=0.)
##    #        Qc    = energy flux via conduction from snow to soil
##    #                (ARHYTHM assumes this to be negligible; Qc=0.)
##    #        Ecc   = cold content of snowpack = amount of energy
##    #                needed before snow can begin to melt [J/m^2]
##
##    #        All Q's here have units of [W/m^2].
##    #        Are they all treated as positive quantities ?
##
##    #        rho_air  = density of air [kg/m^3]
##    #        rho_snow = density of snow [kg/m^3]
##    #        Cp_air   = specific heat of air [J/kg/deg_C]
##    #        Cp_snow  = heat capacity of snow [J/kg/deg_C]
##    #                 = ???????? = specific heat of snow
##    #        Kh       = eddy diffusivity for heat [m^2/s]
##    #        Ke       = eddy diffusivity for water vapor [m^2/s]
##    #        Lv       = latent heat of vaporization [J/kg]
##    #        Lf       = latent heat of fusion [J/kg]
##    #        ------------------------------------------------------
##    #        Dn       = bulk exchange coeff for the conditions of
##    #                   neutral atmospheric stability [m/s]
##    #        Dh       = bulk exchange coeff for heat
##    #        De       = bulk exchange coeff for vapor
##    #        ------------------------------------------------------
##    #        T_air    = air temperature [deg_C]
##    #        T_surf   = surface temperature [deg_C]
##    #        T_snow   = average snow temperature [deg_C]
##    #        RH       = relative humidity [none] (in [0,1])
##    #        e_air    = air vapor pressure at height z [mbar]
##    #        e_surf   = surface vapor pressure [mbar]
##    #        ------------------------------------------------------
##    #        h_snow   = snow depth [m]
##    #        z        = height where wind speed is uz [m]
##    #        uz       = wind speed at height z [m/s]
##    #        P0       = atmospheric pressure [mbar]
##    #        T0       = snow temperature when isothermal [deg_C]
##    #                   (This is usually 0.)
##    #        z0_air   = surface roughness length scale [m]
##    #                   (includes vegetation not covered by snow)
##    #                   (Values from page 1033: 0.0013, 0.02 [m])
##    #        kappa    = von Karman's constant [unitless] = 0.41
##    #        dt       = snowmelt timestep [seconds]
##    #----------------------------------------------------------------
##   
##    # FORWARD_FUNCTION Richardson_Number
##    
##    #---------------------------------
##    #Some required physical constants
##    #are defined in the functions:
##    #e.g. Lv, Lf
##    #---------------------------------
##    
##    #------------------------------
##    #Compute the Richardson number
##    #------------------------------
##    Ri = Richardson_Number(z, uz, T_air, T_surf)
##    
##    #-------------------------------------------------
##    #Compute bulk exchange coeffs (neutral stability)
##    #-------------------------------------------------
##    Dn = Bulk_Exchange_Coeff(uz, z, h_snow, z0_air, T_air, T_surf)
##    Dh = Dn
##    De = Dn
##    
##    #---------------------------
##    #Compute sensible heat flux
##    #---------------------------
##    Qh = Sensible_Heat_Flux(rho_air, Cp_air, Dh, T_air, T_surf)
##    #Formula:  Qh = rho_air * Cp_air * Dh * (T_air - T_surf)
##    #print,'Dh = ', Dh
##    #print,'Qh = ', Qh
##    
##    #-------------------------
##    #Compute latent heat flux
##    #-------------------------
##    Qe = Latent_Heat_Flux(rho_air, De, T_air, T_surf, RH, p0,
##                          e_air, e_surf)  #(these 2 returned)
##    #Formula:  Qe = rho_air * Lv * De * (0.662/p0) * (e_air - e_surf)
##    
##    #print,'Qe = ', Qe
##    
##    #-----------------------------
##    #Compute conduction heat flux
##    #-----------------------------
##    Qc = Conduction_Heat_Flux()
##    #Formula:  Qc = 0d
##    
##    #-----------------------------
##    #Compute advective heat flux
##    #-----------------------------
##    Qa = Advection_Heat_Flux()
##    #Formula:  Qa = 0d
##    
##    #---------------------------------
##    #Qn_SW, Qn_SW & Ecc are pointers,
##    #others are local variables
##    #----------------------------------------------------
##    #Ecc is initialized with the Initial_Cold_Content
##    #function by Initialize_Snow_Vars function (2/21/07)
##    #----------------------------------------------------
##    #The following pseudocode only works for scalars but
##    #is otherwise equivalent to that given below and
##    #clarifies the logic:
##    #----------------------------------------------------
##    #  if (Q_sum gt 0) then begin
##    #      if ((Q_sum * dt) gt Ecc) then begin
##    #          ;-----------------------------------------
##    #          ;Snow is melting.  Use some of Q_sum to
##    #          ;overcome Ecc, and remainder to melt snow
##    #          ;-----------------------------------------
##    #          Qm  = Q_sum - (Ecc/dt)
##    #          Ecc = 0
##    #          M   = (Qm / (rho_w * Lf))
##    #      endif else begin
##    #          ;----------------------------
##    #          ;Snow is warming; reduce Ecc
##    #          ;----------------------------
##    #          Ecc = (Ecc - (Q_sum * dt))
##    #          M   = 0d
##    #      endelse
##    #  endif else begin
##    #      ;------------------------------
##    #      ;Snow is cooling; increase Ecc
##    #      ;------------------------------
##    #      Ecc = Ecc - (Q_sum * dt)
##    #      M   = 0d
##    #  endelse
##    #-------------------------------------------------------
##    Q_sum = Qn_SW + Qn_LW + Qh + Qe + Qa + Qc    #[W/m^2]
##    Qcc  = (Ecc / dt)                            #[W/m^2]
##    Qm   = maximum((Q_sum - Qcc), float64(0))                      #[W/m^2]
##    Ecc  = maximum((Ecc - (Q_sum * dt)), float64(0))              #[J/m^2]
##    #print,'Qm = ', Qm
##    #print,' '
##    
##    #-----------------------------------
##    #Convert melt energy to a melt rate
##    #----------------------------------------
##    #Lf = latent heat of fusion [J/kg]
##    #Lv = latent heat of vaporization [J/kg]
##    #M  = (Qm/ (rho_w * Lf))
##    #----------------------------------------
##    #rho_w = 1000d       ;[kg/m^3]
##    #Lf    = 334000d     ;[J/kg = W*s/kg]
##    #So (rho_w * Lf) = 3.34e+8  [J/m^3]
##    #-------------------------------------
##    M = (Qm / float32(3.34E+8))   #[m/s]
##    
##    return maximum(M, float32(0.0))
##
###   Energy_Balance_Meltrate
###-------------------------------------------------------------------------
